
from collections import namedtuple
from inspect import stack, getmodule
from json import loads
from logging import getLogger
from random import choice
from re import sub, compile
from string import ascii_letters

import pandas as pd
from binascii import hexlify
from sqlalchemy.orm.exc import NoResultFound

from ..stats.names import BOOKMARK

log = getLogger(__name__)


class WarnDict(dict):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            msg = self.__msg % (item,)
            self.__log.debug(msg)
            raise KeyError(msg)


class WarnList(list):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except IndexError:
            msg = self.__msg % item
            self.__log.debug(msg)
            raise IndexError(msg)


def tohex(data):
    return hexlify(data).decode('ascii')


def assert_attr(instance, *attrs):
    for attr in attrs:
        if getattr(instance, attr) is None:
            raise Exception('No %s defined' % attr)


def kargs_to_attr(**kargs):
    return dict_to_attr(kargs)


def dict_to_attr(kargs):
    return namedtuple('Attr', kargs.keys(), rename=True)(*kargs.values())


class MutableAttr(dict):

    def __init__(self, *args, none=False, **kargs):
        self.__none = none
        super().__init__(*args, **kargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            if self.__none:
                return None
            else:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self[name] = value

    def _to_dict(self):
        return self.__dict__


class MissingReference(Exception): pass


def reftuple(name, *args, **kargs):
    '''
    Like a namedtuple, but expands $ strings using a database session and date
    (# is similar, but also does JSON parsing).
    '''

    class klass(namedtuple(name, *args, **kargs)):

        def expand(self, s, time, default_owner=None, default_constraint=None):
            instance = self
            for name in self._fields:
                value = getattr(instance, name)
                log.debug(f'Expanding {name}: {value} at {time}')
                value = expand(s, value, time, default_owner=default_owner, default_constraint=default_constraint)
                log.debug(f'Setting {name} = {value}')
                instance = instance._replace(**{name: value})
            return instance

    klass.__name__ = name
    caller = stack()[1]
    klass.__module__ = getmodule(None, caller.filename).__name__
    return klass


class MaxDict(dict):

    def __init__(self, kv):
        super().__init__()
        for key, value in kv:
            if key in self:
                self[key] = max(value, self[key])
            else:
                self[key] = value


def tmp_name():
    return ''.join(choice(ascii_letters) for _ in range(8))


def nearest_index(df, name, value):
    exactmatch = df.loc[df[name] == value]
    if not exactmatch.empty:
        return exactmatch.index[0]
    else:
        lower = df.loc[df[name] < value].index.dropna()
        upper = df.loc[df[name] > value].index.dropna()
        if lower.empty:
            return upper.min()
        elif upper.empty:
            return lower.max()
        else:
            if abs(value - df.loc[lower.max()][name]) < abs(value - df.loc[upper.min()][name]):
                return lower.max()
            else:
                return upper.min()


def get_index_loc(df, value):
    loc = df.index.get_loc(value)
    try:
        return loc.start  # if slice, take first
    except AttributeError:
        return loc  # otherwise, simple value


def linscale(series, lo=0, hi=None, min=0, max=1, gamma=1):
    lo = series.dropna().min() if lo is None else lo
    hi = series.dropna().max() if hi is None else hi
    return (min + (max - min) * ((series - lo) / (hi - lo)) ** gamma).clip(min, max)


def sorted_numeric_labels(labels, text=None):
    if text:
        labels = [label for label in labels if text in label]
    return sorted(labels, key=lambda label: int(sub(r'\D', '', label)))


def interpolate_freq(df, freq, **kargs):
    left = pd.DataFrame(index=pd.date_range(df.index.min(), df.index.max(), freq=freq))
    return left_interpolate(left, df, **kargs)


def left_interpolate(left, right, **kargs):
    '''
    interpolate right so that it is on the same index as left.
    '''
    # neater solution https://stackoverflow.com/questions/47148446/pandas-resample-interpolate-is-producing-nans
    # option 1 with reindexing
    tmp = tmp_name()
    left[tmp] = True
    both = left.join(right, how='outer').interpolate(**kargs)
    return both.loc[both[tmp] == True].drop(columns=[tmp])


def bookend(df, column=BOOKMARK):
    # https://stackoverflow.com/questions/53927414/get-only-the-first-and-last-rows-of-each-group-with-pandas
    g = df.groupby(column)
    return pd.concat([g.head(1), g.tail(1)]).drop_duplicates().sort_index()


def expand(s, text, before, vars=None, default_owner=None, default_constraint=None):
    '''
    Recursively expand any ${name} occurrences in the text using vars (if given) and database.

    May be too much magic going on here - can return objects, values as well as strings.
    '''

    from ..sql import StatisticName, StatisticJournal

    if vars is None: vars = {}
    pattern = compile(r'(.*)\${([^}]+)}(.*)')

    match = pattern.match(text)
    while match:
        left, name, right = match.groups()
        log.debug(f'Found "{name}" in {text}')
        if name in vars:
            owner = None
            value = vars[name]
        else:
            owner, statistic, constraint = StatisticName.parse(name, default_owner=default_owner,
                                                               default_constraint=default_constraint)
            value = StatisticJournal.before_not_null(s, before, statistic, owner, constraint)
        if value is None:
            raise Exception(f'No value defined for {name} ({owner}:{statistic}:{constraint}) before {before}')
        elif left == '' and right == '':
            text = value.value
            match = None
            if owner == 'Constant':
                text = loads(text)
            log.debug(f'Unpacked {name}={text}')
        else:
            value = str(value.value)
            log.debug(f'Substituting {name}="{value}" in "{text}"')
            text = left + value + right
            match = pattern.match(text)
    return text


def median(list):
    if not list: raise ValueError('Median of empty set')
    list = sorted(list)
    n = len(list)
    if n % 2:
        return list[n//2]
    else:
        return (list[n//2] + list[n//2+1]) / 2




import datetime as dt
from collections import defaultdict
from itertools import groupby
from logging import getLogger
from os.path import exists, join
from zipfile import ZipFile

from shapely.wkt import loads

from ..sql import Constant

log = getLogger(__name__)


MSIL2A_DIR = 'MSIL2A.Dir'


def clean_and_group_products(products):
    products = sorted(products.values(), key=lambda product: product['beginposition'], reverse=True)
    footprint_areas = sorted((loads(product['footprint']).area for product in products), reverse=True)
    footprint_cutoff = footprint_areas[0] * 0.9
    products = [product for product in products if loads(product['footprint']).area > footprint_cutoff]
    log.debug(f'After footprint area cutoff of {footprint_cutoff} have {len(products)} products')
    neighbours = [a['beginposition'] - b['beginposition'] < dt.timedelta(hours=3)
                  for a, b in zip(products, products[1:])]

    def make_counter():
        count = 0

        def counter(neighbour):
            nonlocal count
            if not neighbour: count += 1
            return count

        return counter

    counter = make_counter()
    groups = [0] + [counter(neighbour) for neighbour in neighbours]
    productss = [[pg[0] for pg in pgs] for g, pgs in groupby(zip(products, groups), key=lambda pg: pg[1])]
    max_size = max(*[len(products) for products in productss])
    log.debug(f'Grouped products into {len(productss)} groups, with maximum size {max_size}')
    productss = [products for products in productss if len(products) == max_size]
    log.debug(f'Finally have {len(productss)} maximally-sized groups')
    return productss


def score_productss(productss, dir):
    scored = defaultdict(list)
    for products in productss:
        score, hit, miss = 0, [], []
        for product in products:
            if exists(join(dir, product['uuid'])):
                hit.append(product)
            else:
                score += 1
                miss.append(product)
        scored[score].append((hit, miss))
    scores = sorted(list(scored.keys()))
    log.debug(f'Scores range from {scores[0]} to {scores[-1]}')
    return scored[scores[0]][0]


def check_and_unpack_success(products, success, data_dir):
    products_by_id = {product['uuid']: product for product in products}
    for id in success:
        if id not in products_by_id:
            raise Exception(f'Downloaded unexpected {id}')
        if not success[id]['path']:
            log.debug(success)
            raise Exception(f'No path for {id}')
        src = success[id]['path']
        dest = join(data_dir, id)
        if exists(dest):
            raise Exception(f'Destination {dest} already exists')
        log.debug(f'Extracting {id} to {dest}')
        ZipFile(src).extractall(path=dest)


def download_missing(api, products, data_dir, download_dir='/tmp'):
    '''
    Download to download_dir and unpack in data_dir.
    We do not delete from download_dir - the system will do that for us and the underlying lib will use it as a cache
    if for some reason the unpacked data were deleted.
    '''
    # list comprehension to work round bug in sentinelsat 0.13
    success, lta, failure = api.download_all([product['uuid'] for product in products], directory_path=download_dir)
    log.debug(f'success: {len(success)}; lta: {len(lta)}; failure: {len(failure)}')
    if success:
        check_and_unpack_success(products, success, data_dir)
    if failure:
        raise Exception(f'Failed to download {", ".join(failure)}.')
    if lta:
        raise Exception(f'Some downloads triggered LTA ({", ".join(lta.keys())})')


def cached_download(s, api, products, dir_name=MSIL2A_DIR):
    '''
    The products (a dict) are the result if a query to SentinelAPI.  They describe various candidate files.
    We assume that products at nearby times need to be combined so must be downloaded together.
    This routine tried to find a suitable set of products that have full coverage, can be downloaded
    together, and preferably already exist on disk.
    '''
    productss = clean_and_group_products(products)
    msil2a_dir = Constant.get(s, dir_name).at(s).value
    hit, miss = score_productss(productss, msil2a_dir)
    if miss:
        download_missing(api, miss, msil2a_dir)
    return [join(msil2a_dir, product['uuid']) for product in hit+miss]





class Root:

    def __init__(self, *args, **kargs):
        print('Root init')
        print(self.bar)
        print(f'Root {dir(self)}')
        print(self.bar)


class ChildA(Root):

    def __init__(self, *args, **kargs):
        print('Child A init')
        print(self.bar)
        super().__init__(*args, **kargs)
        print(f'Child A {dir(self)}')
        print(self.bar)

    def foo(self):
        print('Child A foo')
        yield from self.bar()

    def bar(self):
        print('Child A bar')
        raise NotImplementedError()


class ChildB(Root):

    def __init__(self, *args, **kargs):
        print('Child B init')
        print(self.bar)
        super().__init__(*args, **kargs)
        print(f'Child B {dir(self)}')
        print(self.bar)


class Grandchild(ChildA, ChildB):

    def __init__(self, *args, **kargs):
        print('Grandchild init')
        print(self.bar)
        super().__init__(*args, **kargs)
        print(f'Grandchild {dir(self)}')
        print(self.bar)

    def bar(self):
        print('Grandchild bar')
        yield 1


if __name__ == '__main__':
    x = Grandchild()
    print(list(x.foo()))



from collections import defaultdict
from importlib.resources import read_binary
from logging import getLogger
from os.path import split, sep, splitext

from werkzeug import Response


log = getLogger(__name__)


class Static:

    CONTENT_TYPE = defaultdict(lambda: 'text/plain', {
        'js': 'text/javascript',
        'html': 'text/html',
        'css': 'text/css'
    })

    def __init__(self, package):
        if package.startswith('.'):
            self.__package = __name__.rsplit('.', maxsplit=1)[0] + package
        else:
            self.__package = package

    def __call__(self, request, s, path):
        package, file = self.parse_path(path)
        for (extension, encoding) in (('.gz', 'gzip'), ('', None)):
            if not encoding or encoding in request.accept_encodings:
                file_extn = file + extension
                try:
                    log.debug(f'Reading {file_extn} from {package}')
                    response = Response(read_binary(package, file_extn))
                    if encoding:
                        response.content_encoding = encoding
                    self.set_content_type(response, file)
                    return response
                except Exception as e:
                    if encoding:
                        log.debug(f'Encoding {encoding} not supported by server: {e}')
                    else:
                        log.warning(f'Error serving {file}: {e}')
            else:
                log.debug(f'Encoding {encoding} not supported by client')
        raise Exception(f'File not found: {file}')

    def parse_path(self, path):
        package = self.__package
        head, tail = split(path)
        if not tail:
            raise Exception(f'{path} is a directory')
        if tail == '__init__.py':
            raise Exception('Refusing to serve package marker')
        if '.' in head:
            raise Exception(f'Package separators in {head}')
        if head:
            package += '.' + '.'.join(head.split(sep))
        return package, tail

    def set_content_type(self, response, name):
        ext = splitext(name)[1].lower()
        if ext:
            ext = ext[1:]
        response.content_type = self.CONTENT_TYPE[ext]




from tempfile import TemporaryDirectory
from tests import LogTestCase

from sqlalchemy.sql.functions import count

from ch2.commands.activities import activities
from ch2.commands.args import m, V, DEV, mm, FAST, BASE, bootstrap_dir, FORCE
from ch2.commands.constants import constants
from ch2.config.profile.default import default
from ch2.sql.tables.activity import ActivityJournal
from ch2.sql.tables.pipeline import PipelineType
from ch2.sql.tables.statistic import StatisticJournal, StatisticJournalFloat, StatisticName
from ch2.stats.names import RAW_ELEVATION, ELEVATION, ACTIVE_DISTANCE, ACTIVE_TIME
from ch2.stats.pipeline import run_pipeline
from tests import LogTestCase


class TestActivities(LogTestCase):

    def test_activities(self):

        with TemporaryDirectory() as base:

            bootstrap_dir(base, m(V), '5')

            bootstrap_dir(base, m(V), '5', mm(DEV), configurator=default)

            args, sys, db = bootstrap_dir(base, m(V), '5', 'constants', 'set', 'SRTM1.dir',
                                          '/home/andrew/archive/srtm1', mm(FORCE))
            constants(args, sys, db)

            args, sys, db = bootstrap_dir(base, m(V), '5', mm(DEV), 'activities',
                                          'data/test/source/personal/2018-08-27-rec.fit')
            activities(args, sys, db)

            # run('sqlite3 %s ".dump"' % f.name, shell=True)

            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, force=True, start='2018-01-01', n_cpu=1)

            # run('sqlite3 %s ".dump"' % f.name, shell=True)

            with db.session_context() as s:
                n_raw = s.query(count(StatisticJournalFloat.id)). \
                    join(StatisticName). \
                    filter(StatisticName.name == RAW_ELEVATION).scalar()
                self.assertEqual(2099, n_raw)
                n_fix = s.query(count(StatisticJournalFloat.id)). \
                    join(StatisticName). \
                    filter(StatisticName.name == ELEVATION).scalar()
                self.assertEqual(2099, n_fix)
                # WHY does this jump around?
                n = s.query(count(StatisticJournal.id)).scalar()
                # self.assertEqual(50403, n)
                self.assertTrue(n > 30000)
                self.assertTrue(n < 100000)
                journal = s.query(ActivityJournal).one()
                self.assertNotEqual(journal.start, journal.finish)

    def test_segment_bug(self):
        with TemporaryDirectory() as f:
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            paths = ['/home/andrew/archive/fit/bike/cotic/2016-07-27-pm-z4.fit']
            run_pipeline(sys, db, args[BASE], PipelineType.ACTIVITY, paths=paths, force=True)

    def __assert_basic_stats(self, s):
        for name in [ACTIVE_DISTANCE, ACTIVE_TIME]:
            stat = s.query(StatisticJournal). \
                join(StatisticName). \
                filter(StatisticName.name == name).one_or_none()
            print(f'{name} = {stat}')
            self.assertTrue(stat, f'No value for {name}')

    def test_florian(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                          'data/test/source/private/florian.fit')
            activities(args, sys, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                self.__assert_basic_stats(s)

    def test_michael(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args,sys,  db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                      'data/test/source/other/2019-05-09-051352-Running-iWatchSeries3.fit')
            activities(args, sys, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                self.__assert_basic_stats(s)

    def test_heart_alarms(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                      'data/test/source/personal/2016-07-19-mpu-s-z2.fit')
            activities(args, sys, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                stat = s.query(StatisticJournal). \
                    join(StatisticName). \
                    filter(StatisticName.name == ACTIVE_DISTANCE).one()
                self.assertGreater(stat.value, 30)

    def test_920(self):
        for src in '920xt-2019-05-16_19-42-54.fit', '920xt-2019-05-16_19-42-54.fit':
            with TemporaryDirectory() as f:
                bootstrap_dir(f, m(V), '5')
                bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
                args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                               f'data/test/source/other/{src}')
                activities(args, sys, db)
                # run('sqlite3 %s ".dump"' % f.name, shell=True)
                run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, n_cpu=1)
                # run('sqlite3 %s ".dump"' % f.name, shell=True)
                with db.session_context() as s:
                    self.__assert_basic_stats(s)




from logging import getLogger

from .args import FORCE, LIKE, FINISH, START, WORKER, parse_pairs, KARG, UNLIKE, BASE
from ..sql.tables.pipeline import PipelineType
from ..stats.pipeline import run_pipeline

log = getLogger(__name__)


def statistics(args, sys, db):
    '''
## statistics

    > ch2 statistics

Generate any missing statistics.

    > ch2 statistics --force [DATE]

Delete statistics after the date (or all, if omitted) and then generate new values.
    '''
    run_statistic_pipelines(sys, db, args[BASE], force=args[FORCE], like=args[LIKE], unlike=args[UNLIKE],
                            start=args[START], finish=args[FINISH], worker=args[WORKER] is not None, id=args[WORKER],
                            **parse_pairs(args[KARG]))


def run_statistic_pipelines(sys, db, base, force=False, like=tuple(), unlike=tuple(), start=None, finish=None,
                            worker=False, id=None, **kargs):
    run_pipeline(sys, db, base, PipelineType.STATISTIC, force=force, like=like, unlike=unlike, start=start, finish=finish,
                 worker=worker, id=id, **kargs)



from glob import glob
from logging import getLogger
from os.path import dirname, basename, join

from sqlalchemy.orm.exc import NoResultFound

from ...commands.args import DB_VERSION
from ...lib import format_date, time_to_local_date, to_time
from ...lib.utils import clean_path
from ...sql import StatisticJournal, StatisticName, StatisticJournalType
from ...sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ...sql.types import short_cls
from ...sql.utils import add

log = getLogger(__name__)


def journal_imported(record, new, cls, name):
    # true if already installed
    with new.session_context() as new_s:
        if new_s.query(StatisticJournal). \
                join(cls). \
                filter(StatisticJournal.source_id == cls.id). \
                count():
            record.warning(f'{name} topic entries already exist - old data must be imported first')
            return True
    return False


def match_statistic_name(record, old_statistic_name, new_s, owner, constraint):
    try:
        log.debug(f'Trying to find new statistic_name for {old_statistic_name}')
        new_statistic_name = new_s.query(StatisticName). \
            filter(StatisticName.name == old_statistic_name.name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint,
                   StatisticName.statistic_journal_type == old_statistic_name.statistic_journal_type).one()
        log.debug(f'Found new statistic_name {new_statistic_name}')
        return new_statistic_name
    except NoResultFound:
        record.raise_(f'No new equivalent to statistic {old_statistic_name.name} '
                      f'({StatisticJournalType(old_statistic_name.statistic_journal_type).name}) '
                      f'for {short_cls(owner)} / {constraint}')


def copy_statistic_journal(record, old_s, old, old_statistic_name, old_statistic_journal,
                           new_s, new_statistic_name, source, name=None):
    journals = {StatisticJournalType.INTEGER.value: old.meta.tables['statistic_journal_integer'],
                StatisticJournalType.FLOAT.value: old.meta.tables['statistic_journal_float'],
                StatisticJournalType.TEXT.value: old.meta.tables['statistic_journal_text']}
    journal = journals[old_statistic_name.statistic_journal_type]
    old_value = old_s.query(journal).filter(journal.c.id == old_statistic_journal.id).one()
    log.debug(f'Resolved old statistic_journal {old_value}')
    new_value = add(new_s,
                    STATISTIC_JOURNAL_CLASSES[StatisticJournalType(new_statistic_name.statistic_journal_type)](
                        value=old_value.value, time=old_statistic_journal.time, statistic_name=new_statistic_name,
                        source=source))
    new_s.commit()  # avoid logging below if error
    date = format_date(time_to_local_date(to_time(new_value.time)))
    name = name if name else new_statistic_name.name
    record.info(f'Statistic {new_value.value} at {date} for {name}')


def any_attr(instance, *names):
    log.debug(dir(instance))
    for name in names:
        if hasattr(instance, name):
            return getattr(instance, name)
    raise AttributeError(f'No {names} in {instance} ({type(instance)})')


def available_versions(base):
    versions = []
    base = clean_path(base)
    if base.endswith(DB_VERSION): base = dirname(base)
    log.debug(f'Looking for previous versions under {base}')
    append(versions, (basename(candidate) for candidate in glob(join(base, '[0-9]-[0-9]*'))
                      if basename(candidate) != DB_VERSION))
    append(versions, glob(join(base, '**/database-[0-9]*-[0-9]*.sql'), recursive=True))
    append(versions, glob(join(base, '**/database-[0-9]*-[0-9]*.db'), recursive=True))
    return versions


def append(versions, glob):
    for version in sorted(glob, reverse=True):
        log.debug(version)
        versions.append(version)



from json import loads
from logging import getLogger

from ...commands.args import base_system_path
from ...commands.configure import load, delete
from ...commands.help import HTML, filter, parse, P, LI, PRE
from ...commands.upgrade import upgrade_path
from ...lib.log import Record
from ...config.utils import profiles
from ...lib import time_to_local_time, local_time_to_time
from ...lib.utils import restart_self
from ...migrate.import_ import available_versions
from ...migrate.import_.activity import activity_imported
from ...migrate.import_.constant import constant_imported
from ...migrate.import_.diary import diary_imported
from ...migrate.import_.kit import kit_imported
from ...sql import SystemConstant, Constant, StatisticJournal

log = getLogger(__name__)


ACTIVITY = 'activity'
CONFIGURED = 'configured'
CONSTANT = 'constant'
DESCRIPTION = 'description'
DIARY = 'diary'
DIRECTORY = 'directory'
IMPORTED = 'imported'
KIT = 'kit'
NAME = 'name'
PROFILE = 'profile'
PROFILES = 'profiles'
SINGLE = 'single'
STATISTIC = 'statistic'
TIME = 'time'
VALUE = 'value'
VALUES = 'values'
VERSION = 'version'
VERSIONS = 'versions'


class Configure:

    def __init__(self, sys, db, base):
        self.__sys = sys
        self.__db = db
        self.__base = base
        self.__html = HTML(delta=1, parser=filter(parse, yes=(P, LI, PRE)))

    def is_configured(self):
        return bool(self.__sys.get_constant(SystemConstant.DB_VERSION, none=True))

    def html(self, text):
        return self.__html.str(text)

    def read_profiles(self, request, s):
        fn_argspec_by_name = profiles()
        version = self.__sys.get_constant(SystemConstant.DB_VERSION, none=True)
        data = {PROFILES: {name: self.html(fn_argspec_by_name[name][0].__doc__) for name in fn_argspec_by_name},
                CONFIGURED: bool(version),
                DIRECTORY: self.__base}
        if data[CONFIGURED]: data[VERSION] = version
        return data

    def write_profile(self, request, s):
        data = request.json
        load(self.__sys, s, self.__base, False, data[PROFILE])

    def delete(self, request, s):
        delete(self.__sys, base_system_path(self.__base), True)
        # now we need to restart because the database connections exist
        restart_self()

    def read_upgrade(self, request, s):
        record = Record(log)
        return {IMPORTED: {DIARY: diary_imported(record, self.__db),
                           ACTIVITY: activity_imported(record, self.__db),
                           KIT: kit_imported(record, self.__db),
                           CONSTANT: constant_imported(record, self.__db)},
                VERSIONS: available_versions(self.__base)}

    def write_upgrade(self, request, s):
        data = request.json
        record = Record(log)
        upgrade_path(record, self.__base, data[VERSION], self.__db)
        return record.json()

    def read_constants(self, request, s):
        return [self.read_constant(s, constant)
                for constant in s.query(Constant).order_by(Constant.name).all()]

    def read_constant(self, s, constant):
        as_json = bool(constant.validate_cls)
        values = [{TIME: time_to_local_time(statistic.time),
                   VALUE: loads(statistic.value) if as_json else statistic.value,
                   STATISTIC: statistic.id}
                  for statistic in s.query(StatisticJournal).
                      filter(StatisticJournal.source_id == constant.id).all()]
        return {NAME: constant.name,
                SINGLE: constant.single,
                DESCRIPTION: self.html(constant.statistic_name.description),
                VALUES: values}

    def write_constant(self, request, s):
        data = request.json
        log.debug(data)
        constant = Constant.get(s, data[NAME])
        value = data[VALUES][0][VALUE]
        time = data[VALUES][0][TIME]
        statistic_journal_id = data[VALUES][0][STATISTIC]
        if statistic_journal_id:
            journal = s.query(StatisticJournal). \
                filter(StatisticJournal.id == statistic_journal_id,
                       StatisticJournal.source_id == constant.id).one()
            journal.set(value)
            if not constant.single:
                journal.time = local_time_to_time(time)
        else:
            if constant.single and time:
                log.warning(f'Ignoring time for {constant.name}')
                time = 0.0
            constant.add_value(s, value, time=time)




from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np

from . import MultiProcCalculator, DataFrameCalculatorMixin, ActivityGroupCalculatorMixin
from ..names import FTHR, HEART_RATE, HR_ZONE, ALL, HR_IMPULSE_10, SUM
from ...data.frame import activity_statistics, statistics
from ...data.impulse import hr_zone, impulse_10
from ...sql import Constant, StatisticJournalFloat, ActivityGroup

log = getLogger(__name__)

# constraint comes from constant
HRImpulse = namedtuple('HRImpulse', 'gamma, zero, one, max_secs')


class ImpulseCalculator(ActivityGroupCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, impulse_ref=None, **kargs):
        self.impulse_ref = self._assert('impulse_ref', impulse_ref)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        self.impulse = HRImpulse(**loads(Constant.get(s, self.impulse_ref).at(s).value))
        self.all = ActivityGroup.from_name(s, ALL)
        log.debug('%s: %s' % (self.impulse_ref, self.impulse))

    def _read_dataframe(self, s, ajournal):
        try:
            heart_rate_df = activity_statistics(s, HEART_RATE, activity_journal=ajournal)
            fthr_df = statistics(s, FTHR, constraint=ajournal.activity_group)
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise
        if fthr_df.empty:
            raise Exception(f'No {FTHR} defined for {ajournal.activity_group}')
        return heart_rate_df, fthr_df

    def _calculate_stats(self, s, ajournal, data):
        heart_rate_df, fthr_df = data
        hr_zone(heart_rate_df, fthr_df)
        impulse_df = impulse_10(heart_rate_df, self.impulse)
        # join so that we can iterate over values in time order
        stats = impulse_df.join(heart_rate_df, how='outer')
        return stats

    def _copy_results(self, s, ajournal, loader, stats):
        for time, row in stats.iterrows():
            if not np.isnan(row[HR_ZONE]):
                loader.add(HR_ZONE, None, None, ajournal.activity_group, ajournal, row[HR_ZONE], time,
                           StatisticJournalFloat)
            if not np.isnan(row[HR_IMPULSE_10]):
                # load a copy to the activity group as well as to all so that we can extract / display
                # easily in, for example, std_activity_statistics
                loader.add(HR_IMPULSE_10, None, None, ajournal.activity_group, ajournal,
                           row[HR_IMPULSE_10], time, StatisticJournalFloat)
                # copy for global FF statistics
                loader.add(HR_IMPULSE_10, None, None, self.all, ajournal,
                           row[HR_IMPULSE_10], time, StatisticJournalFloat)
        # if there are no values, add a single null so we don't re-process
        if not loader:
            loader.add(HR_ZONE, None, SUM, ajournal.activity_group, ajournal, None, ajournal.start,
                       StatisticJournalFloat)




from logging import getLogger

from . import template
from ..lib.inspect import read_package

log = getLogger(__name__)


def templates():
    return dict(read_package(template))


def get_template(name):
    return templates()[name]



from logging import getLogger
from os.path import sep, exists, join, isfile

from .args import SOURCE, ACTIVITY, DB_EXTN, base_system_path
from ..lib.utils import clean_path
from ..lib.log import Record
from ..migrate.import_.activity import upgrade_activity
from ..migrate.import_.constant import upgrade_constant
from ..migrate.import_.diary import upgrade_diary
from ..migrate.import_.kit import upgrade_kit
from ..sql.database import ReflectedDatabase

log = getLogger(__name__)


def upgrade(args, sys, db):
    '''
## import

    > ch2 import 0-30

Import diary entries from a previous version.
    '''
    upgrade_path(Record(log), args, args[SOURCE], db)


def upgrade_path(record, base, source, new):
    path = build_source_path(record, base, source)
    old = ReflectedDatabase(path, read_only=True)
    if not old.meta.tables:
        record.raise_(f'No tables found in {path}')
    log.info(f'Importing data from {path}')
    upgrade_diary(record, old, new)
    upgrade_activity(record, old, new)
    upgrade_kit(record, old, new)
    upgrade_constant(record, old, new)


def build_source_path(record, base, source):

    def nice_msg(template, source, path):
        msg = template
        if source != path: msg += f' ({path})'
        return msg

    database = ACTIVITY + DB_EXTN
    if sep not in source:
        path = base_system_path(base, file=database, version=source, create=False)
        if exists(path):
            log.info(nice_msg(f'{source} appears to be a version', source, path))
            return path
        else:
            log.warning(nice_msg(f'{source} is not a version', source, path))
    path = clean_path(source)
    if exists(path) and isfile(path):
        log.info(nice_msg(f'{source} exists', source, path))
        return path
    else:
        log.warning(f'{source} is not a database file ({path})')
    path = join(path, database)
    if exists(path) and isfile(path):
        log.info(nice_msg(f'{source} exists', source, path))
        return path
    else:
        log.warning(nice_msg(f'{source} is not a base directory', source, path))
    record.raise_(f'Could not find {source}')






from ..config import Config, WALK, SWIM, RUN, BIKE
from ..database import add_diary_topic, add_child_diary_topic, add_diary_topic_field, add_nearby, add_enum_constant, \
    add_constant
from ..power import add_power_estimate
from ...commands.args import DEFAULT, base_system_path, PERMANENT
from ...diary.model import TYPE, EDIT
from ...lib import to_time, time_to_local_date
from ...msil2a.download import MSIL2A_DIR
from ...sql import StatisticJournalType, StatisticName, DiaryTopic, DiaryTopicJournal
from ...sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ...sql.utils import add
from ...stats.calculate.power import Bike
from ...stats.names import SPORT_CYCLING, SPORT_RUNNING, SPORT_SWIMMING, SPORT_WALKING


def acooke(sys, s, base, no_diary=False):
    '''
## acooke

This extends the default configuration with:

* Diary entries that I need
* Additional activity groups selected on the kit used
* Power estimates
* An area around Santiago, Chile, for registering nearby activities

Unlikely to be useful to others, but works as an example of how you can extend the code yourself.
    '''
    ACooke(sys, base, no_diary=no_diary).load(s)


ROAD = 'Road'
MTB = 'MTB'


class ACooke(Config):

    def _load_diary_topics(self, s, c):
        super()._load_diary_topics(s, c)

        # add more diary entries specific to my needs

        injuries = add_diary_topic(s, 'Injuries', c)

        ms = add_child_diary_topic(s, injuries, 'Multiple Sclerosis', c)
        add_diary_topic_field(s, ms, 'Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, ms, 'Betaferon', c,
                              schedule='2018-08-07/2d[1]')  # reminder to take meds on alternate days

        leg = add_child_diary_topic(s, injuries, 'Broken Femur LHS', c,
                                    schedule='d2018-03-11-2020-03-01')
        add_diary_topic_field(s, leg, 'Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, leg, 'Learn to manage tendon pain', c)  # aims added as child topics
        add_child_diary_topic(s, leg, 'Maintain fitness', c)
        add_child_diary_topic(s, leg, 'Visit UK', c, schedule='-2018-08-11')

    def _load_specific_activity_groups(self, s):
        super()._load_specific_activity_groups(s)

        # additional activity groups for different cycling activities

        self._load_activity_group(s, ROAD, 'Road cycling activities')
        self._load_activity_group(s, MTB, 'MTB cycling activities')

    def _sport_to_activity(self):

        # map the additional groups above based on kit use
        # (cotic and bowman are kit items added via kit commands)

        return {SPORT_CYCLING: {
                    'kit': {
                        'cotic': MTB,
                        'bowman': ROAD,
                    },
                    DEFAULT: BIKE,
                },
                SPORT_RUNNING: RUN,
                SPORT_SWIMMING: SWIM,
                SPORT_WALKING: WALK}

    def _load_power_statistics(self, s, c):
        # add power estimates for the two bikes
        # (note that this comes after standard stats, but before summary, achievements, etc).
        for name in (MTB, ROAD):
            add_power_estimate(s, c, self._activity_groups[name], vary='')
        for (name, value) in (('Power.cotic', {'cda': 0.42, 'crr': 0.0055, 'weight': 12}),
                              ('Power.bowman', {'cda': 0.42, 'crr': 0.0055, 'weight': 8})):
            add_enum_constant(s, name, Bike, value, single=True, description='''
Parameters to calculate power when using this bike.

The parameter name must match the kit name (see the PowerEstimate constants). 
* Cda is the product of coefficient of drag and frontal area (units m2).
* Crr is the coefficient of rolling resistance.
* Weight is the bike weight (kg).
''')

    def _load_statistics_pipeline(self, s, c):
        super()._load_statistics_pipeline(s, c)
        # define spatial regions for nearby routes etc
        for name in (MTB, ROAD, WALK):
            add_nearby(s, c, self._activity_groups[name], 'Santiago', -33.4, -70.4, fraction=0.1, border=150)

    def _load_constants(self, s):
        super()._load_constants(s)
        add_constant(s, MSIL2A_DIR, base_system_path(self._base, version=PERMANENT, subdir='msil2a', create=False),
                     description='''
Directory containing Sentinel 2A imaging data (see https://scihub.copernicus.eu/dhus/#/home)

This can be used to generate background images for plots.
I used the data in an experiment to generate 3D plots, but it wasn't very successful.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)

    def _post(self, s):
        super()._post(s)
        # set a default weight for early power calculations
        weight = s.query(StatisticName).filter(StatisticName.name == 'Weight', StatisticName.owner == DiaryTopic).one()
        diary = add(s, DiaryTopicJournal(date=time_to_local_date(to_time(0.0))))
        add(s, STATISTIC_JOURNAL_CLASSES[weight.statistic_journal_type](
            value=65.0, time=0.0, statistic_name=weight, source=diary))




from abc import abstractmethod
from logging import getLogger

from sqlalchemy import not_
from sqlalchemy.sql.functions import count

from ..pipeline import MultiProcPipeline, UniProcPipeline, LoaderMixin
from ...commands.args import STATISTICS, WORKER, mm, VERBOSITY, LOG
from ...lib.date import local_time_to_time, time_to_local_time, format_date, to_date
from ...lib.log import log_current_exception
from ...lib.schedule import Schedule
from ...sql import ActivityJournal, Interval, Timestamp, StatisticJournal, StatisticName, SegmentJournal, ActivityGroup
from ...sql.types import long_cls
from ...sql.utils import add

log = getLogger(__name__)


class CalculatorMixin:

    def __init__(self, *args, owner_in=None, start=None, finish=None, **kargs):
        self.owner_in = self._assert('owner_in', owner_in)  # todo - not needed by some...
        self.start = start  # optional start local time (always present for workers)
        self.finish = finish  # optional finish local time (always present for workers)
        super().__init__(*args, **kargs)

    def _start_finish(self, type=None):
        start, finish = self.start, self.finish
        if type:
            if start: start = type(start)
            if finish: finish = type(finish)
        return start, finish

    def _base_command(self):
        return f'{STATISTICS}'


class MultiProcCalculator(CalculatorMixin, MultiProcPipeline):

    pass


class UniProcCalculator(CalculatorMixin, UniProcPipeline):

    def _base_command(self):
        raise Exception('UniProc does not support workers')

    def _args(self, missing, start, finish):
        raise Exception('UniProc does not support workers')


class JournalCalculatorMixin:
    '''
    auto-detects missing entries, deletes on forcing, and schedules threads via owner_out.

    provides access to the journal via _get_source.
    '''

    _journal_type = None

    def _delimit_query(self, q):
        start, finish = self._start_finish(type=local_time_to_time)
        log.debug(f'Delimit times: {start} - {finish}')
        if start:
            q = q.filter(self._journal_type.start >= start)
        if finish:
            q = q.filter(self._journal_type.start <= finish)
        return q

    def _missing(self, s):
        existing_ids = s.query(Timestamp.source_id).filter(Timestamp.owner == self.owner_out)
        q = s.query(self._journal_type.start). \
            filter(not_(self._journal_type.id.in_(existing_ids.cte()))). \
            order_by(self._journal_type.start)
        return [row[0] for row in self._delimit_query(q)]

    def _args(self, missing, start, finish):
        s, f = time_to_local_time(missing[start]), time_to_local_time(missing[finish])
        log.info(f'Starting worker for {s} - {f}')
        return f'"{s}" "{f}"'

    def _delete(self, s):
        start, finish = self._start_finish(type=local_time_to_time)
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        statistic_names = s.query(StatisticName.id).filter(StatisticName.owner == self.owner_out)
        activity_journals = self._delimit_query(s.query(self._journal_type.id))
        statistic_journals = s.query(StatisticJournal.id). \
            filter(StatisticJournal.statistic_name_id.in_(statistic_names.cte()),
                   StatisticJournal.source_id.in_(activity_journals))
        for repeat in range(2):
            if repeat:
                s.query(StatisticJournal).filter(StatisticJournal.id.in_(statistic_journals.cte())). \
                    delete(synchronize_session=False)
                Timestamp.clear_keys(s, activity_journals.cte(), self.owner_out, constraint=None)
            else:
                n = s.query(count(StatisticJournal.id)). \
                    filter(StatisticJournal.id.in_(statistic_journals.cte())).scalar()
                if n:
                    log.warning(f'Deleting {n} statistics for {long_cls(self.owner_out)} from {start} to {finish}')
                else:
                    log.warning(f'No statistics to delete for {long_cls(self.owner_out)} from {start} to {finish}')
        s.commit()

    def _get_source(self, s, time):
        return s.query(self._journal_type).filter(self._journal_type.start == time).one()


class ActivityJournalCalculatorMixin(JournalCalculatorMixin):

    _journal_type = ActivityJournal


class ActivityGroupCalculatorMixin(ActivityJournalCalculatorMixin):

    def __init__(self, *args, activity_group_name=None, **kargs):
        super().__init__(*args, **kargs)
        self.activity_group_name = activity_group_name

    def _missing(self, s):
        existing_ids = s.query(Timestamp.source_id).filter(Timestamp.owner == self.owner_out)
        q = s.query(self._journal_type.start). \
            filter(not_(self._journal_type.id.in_(existing_ids.cte()))). \
            order_by(self._journal_type.start)
        if self.activity_group_name:
            q = q.join(ActivityGroup).filter(ActivityGroup.name == self.activity_group_name)
        return [row[0] for row in self._delimit_query(q)]


class SegmentJournalCalculatorMixin(JournalCalculatorMixin):

    _journal_type = SegmentJournal


class DataFrameCalculatorMixin(LoaderMixin):

    def _run_one(self, s, time_or_date):
        source = self._get_source(s, time_or_date)
        with Timestamp(owner=self.owner_out, source=source).on_success(s):
            try:
                # data may be structured (doesn't have to be simply a dataframe)
                data = self._read_dataframe(s, source)
                stats = self._calculate_stats(s, source, data)
                loader = self._get_loader(s)
                self._copy_results(s, source, loader, stats)
                loader.load()
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}: {e}')
                log_current_exception()

    @abstractmethod
    def _get_source(self, s, time_or_date):
        raise NotImplementedError()

    @abstractmethod
    def _read_dataframe(self, s, source):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_stats(self, s, source, data):
        raise NotImplementedError()

    @abstractmethod
    def _copy_results(self, s, source, loader, stats):
        raise NotImplementedError()


class DirectCalculatorMixin(LoaderMixin):

    def _run_one(self, s, time_or_date):
        source = self._get_source(s, time_or_date)
        with Timestamp(owner=self.owner_out, source=source).on_success(s):
            try:
                data = self._read_data(s, source)
                loader = self._get_loader(s)
                self._calculate_results(s, source, data, loader)
                loader.load()
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}: {e}')
                log_current_exception()

    @abstractmethod
    def _get_source(self, s, time_or_date):
        raise NotImplementedError()

    @abstractmethod
    def _read_data(self, s, source):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, source, data, loader):
        raise NotImplementedError()


class IntervalCalculatorMixin(LoaderMixin):

    def __init__(self, *args, schedule='m', **kargs):
        self.schedule = Schedule(self._assert('schedule', schedule))
        super().__init__(*args, **kargs)

    def _missing(self, s):
        start, finish = self._start_finish(type=to_date)
        return list(Interval.missing_dates(s, self.schedule, self.owner_out, start=start, finish=finish))

    def _args(self, missing, start, finish):
        s, f = format_date(missing[start][0]), format_date(missing[finish][1])
        log.info(f'Starting worker for {s} - {f}')
        return f'"{s}" "{f}"'

    def _delete(self, s):
        start, finish = self._start_finish()
        # we delete the intervals that the statistics depend on and they will cascade
        for repeat in range(2):
            if repeat:
                q = s.query(Interval)
            else:
                q = s.query(count(Interval.id))
            q = q.filter(Interval.schedule == self.schedule,
                         Interval.owner == self.owner_out)
            if start:
                q = q.filter(Interval.finish > start)
            if finish:
                q = q.filter(Interval.start < finish)
            if repeat:
                for interval in q.all():
                    log.debug('Deleting %s' % interval)
                    s.delete(interval)
            else:
                n = q.scalar()
                if n:
                    log.warning('Deleting %d intervals' % n)
                else:
                    log.warning('No intervals to delete')
        s.commit()

    def _run_one(self, s, missing):
        if s.query(Interval). \
            filter(Interval.schedule == self.schedule,
                   Interval.owner == self.owner_out,
                   Interval.start == missing[0]).one_or_none():
            raise Exception('Interval already exists')
        interval = add(s, Interval(schedule=self.schedule, owner=self.owner_out,
                                   start=missing[0], finish=missing[1]))
        s.commit()
        try:
            data = self._read_data(s, interval)
            loader = self._get_loader(s, add_serial=False, clear_timestamp=False)
            self._calculate_results(s, interval, data, loader)
            loader.load()
        except Exception as e:
            log.warning(f'No statistics for {missing} due to error ({e})')
            log_current_exception()

    @abstractmethod
    def _read_data(self, s, interval):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, interval, data, loader):
        raise NotImplementedError()




from logging import getLogger

from sqlalchemy import or_
from sqlalchemy.sql.functions import count

from .model import from_field, text, optional_text, link, value, trim_no_stats
from ..lib import format_date, time_to_local_time
from ..lib.date import YMD, to_time, YMD_HM, HM
from ..sql import DiaryTopic, DiaryTopicJournal, ActivityJournal, StatisticJournal, Interval
from ..stats.calculate.summary import SummaryCalculator
from ..stats.display import read_pipeline
from ..stats.display.nearby import fmt_nearby, nearby_any_time

log = getLogger(__name__)

COMPARE_LINKS = 'compare-links'


def read_date(s, date):
    yield text(date.strftime('%Y-%m-%d - %A'), tag='title')
    topics = list(read_date_diary_topics(s, date))
    if topics: yield topics
    yield from read_pipeline(s, date)
    gui = list(read_gui(s, date))
    if gui: yield gui
    yield list(read_database_health(s, date))


@optional_text('Diary')
def read_date_diary_topics(s, date):
    journal = DiaryTopicJournal.get_or_add(s, date)
    for topic in s.query(DiaryTopic).filter(DiaryTopic.parent == None,
                                            or_(DiaryTopic.start <= date, DiaryTopic.start == None),
                                            or_(DiaryTopic.finish >= date, DiaryTopic.finish == None)). \
            order_by(DiaryTopic.sort).all():
        if topic.schedule.at_location(date):
            yield list(read_date_diary_topic(s, date, journal.cache(s), topic))


def read_date_diary_topic(s, date, cache, topic):
    yield text(topic.name)
    if topic.description: yield text(topic.description)
    log.debug(f'topic id {topic.id}; fields {topic.fields}')
    for field in topic.fields:
        if field.schedule.at_location(date):
            yield from_field(field, cache[field])
    for child in topic.children:
        if child.schedule.at_location(date):
            content = list(read_date_diary_topic(s, date, cache, child))
            if content:
                # single entries are just text fields
                if len(content) == 1:
                    yield content[0]
                else:
                    yield content


@optional_text('Jupyter')
def read_gui(s, date):
    for aj1 in ActivityJournal.at_date(s, date):
        yield list(read_activity_gui(s, aj1))
    yield link('Health', db=(format_date(date),))


def read_activity_gui(s, aj1):
    title = '%s - %s' % (time_to_local_time(to_time(aj1.start), fmt=YMD_HM),
                         time_to_local_time(to_time(aj1.finish), fmt=HM))
    yield text(title, tag='jupyter-activity')
    links = [link('None', db=(time_to_local_time(aj1.start), None, aj1.activity_group.name))] + \
            [link(fmt_nearby(aj2, nb),
                  db=(time_to_local_time(aj1.start), time_to_local_time(aj2.start), aj1.activity_group.name))
             for aj2, nb in nearby_any_time(s, aj1)]
    yield [text('Compare to', tag=COMPARE_LINKS)] + links
    yield link('All Similar', db=(time_to_local_time(aj1.start), aj1.activity_group.name))


def read_schedule(s, schedule, date):
    yield text(date.strftime(YMD) + ' - Summary for %s' % schedule.describe(), tag='title')
    topics = list(read_schedule_topics(s, schedule, date))
    if topics: yield topics
    yield from read_pipeline(s, date, schedule=schedule)
    gui = list(read_schedule_gui(s, schedule, date))
    if gui: yield gui


@optional_text('Diary')
@trim_no_stats
def read_schedule_topics(s, schedule, start):
    finish = schedule.next_frame(start)
    for topic in s.query(DiaryTopic).filter(DiaryTopic.parent == None,
                                            or_(DiaryTopic.start < finish, DiaryTopic.start == None),
                                            or_(DiaryTopic.finish >= start, DiaryTopic.finish == None)). \
            order_by(DiaryTopic.sort).all():
        yield list(read_schedule_topic(s, schedule, start, finish, topic))


def read_schedule_topic(s, schedule, start, finish, topic):
    yield text(topic.name)
    if topic.description: yield text(topic.description)
    for field in topic.fields:
        column = list(summary_column(s, schedule, start, field.statistic_name))
        if column: yield column
    for child in topic.children:
        if (child.start is None or child.start < finish) and (child.finish is None or child.finish > start):
            content = list(read_schedule_topic(s, schedule, start, finish, child))
            if content: yield content


def summary_column(s, schedule, start, name):
    journals = StatisticJournal.at_interval(s, start, schedule, SummaryCalculator, name, SummaryCalculator)
    for named, journal in enumerate(journal for journal in journals if journal.value != 0):
        summary, period, name = SummaryCalculator.parse_name(journal.statistic_name.name)
        if not named:
            yield text(name)
        yield value(summary, journal.value, units=journal.statistic_name.units)


@optional_text('Jupyter')
def read_schedule_gui(s, schedule, start):
    finish = schedule.next_frame(start)
    yield link('All Activities', db=(format_date(start), format_date(finish)))


@optional_text('Database')
def read_database_health(s, date):
    total = s.query(count(Interval.id)).filter(Interval.dirty == True).scalar()
    today = s.query(count(Interval.id)). \
        filter(Interval.dirty == True,
               Interval.start <= date,
               Interval.finish > date).scalar()
    if not total:
        yield text('No dirty statistics')
    else:
        if today:
            yield value('Dirty statistics today', today)
        yield value('Total dirty statistics', total)
        yield text('Recalculate statistics')




import datetime as dt

from IPython.core.display import display
from bokeh.io import output_file
from bokeh.layouts import row, column, gridplot
from bokeh.plotting import show

from ch2.data import *
from ch2.jupyter.decorator import template
from ch2.lib import *


@template
def activity_details(local_time, activity_group_name):

    f'''
    # Activity Details: {local_time} ({activity_group_name})
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')

    activity = std_activity_statistics(s, local_time=local_time, activity_group_name=activity_group_name)
    details = activity_statistics(s, 'Climb %', ACTIVE_TIME, ACTIVE_DISTANCE, local_time=local_time,
                                  activity_group_name=activity_group_name)
    health = std_health_statistics(s)
    hr_zones = hr_zones_from_database(s, local_time, activity_group_name)

    f'''
    ## Activity Plots
    
    To the right of each plot of data against distance is a related plot of cumulative data
    (except the last, cadence, which isn't useful and so replaced by HR zones).
    Green and red areas indicate differences between the two dates. 
    Additional red lines on the altitude plot are auto-detected climbs.
    
    Plot tools support zoom, dragging, etc.
    '''

    output_file(filename='/dev/null')

    sp = comparison_line_plot(700, 200, DISTANCE_KM, MED_SPEED_KMH, activity, ylo=0)
    add_climb_zones(sp, details, activity)
    sp_c = cumulative_plot(200, 200, MED_SPEED_KMH, activity, ylo=0)

    el = comparison_line_plot(700, 200, DISTANCE_KM, ELEVATION_M, activity, x_range=sp.x_range)
    add_climbs(el, details, activity)
    el_c = cumulative_plot(200, 200, CLIMB_MS, activity)

    hri = comparison_line_plot(700, 200, DISTANCE_KM, HR_IMPULSE_10, activity, ylo=0, x_range=sp.x_range)
    add_climb_zones(hri, details, activity)
    hri_c = cumulative_plot(200, 200, HR_IMPULSE_10, activity, ylo=0)

    hr = comparison_line_plot(700, 200, DISTANCE_KM, HEART_RATE_BPM, activity, x_range=sp.x_range)
    add_hr_zones(hr, activity, DISTANCE_KM, hr_zones)
    add_climb_zones(hr, details, activity)
    hr_c = cumulative_plot(200, 200, HEART_RATE_BPM, activity)

    pw = comparison_line_plot(700, 200, DISTANCE_KM, MED_POWER_ESTIMATE_W, activity, ylo=0, x_range=sp.x_range)
    add_climb_zones(pw, details, activity)
    pw_c = cumulative_plot(200, 200, MED_POWER_ESTIMATE_W, activity, ylo=0)

    cd = comparison_line_plot(700, 200, DISTANCE_KM, MED_CADENCE, activity, ylo=0, x_range=sp.x_range)
    add_climb_zones(cd, details, activity)
    hr_h = histogram_plot(200, 200, HR_ZONE, activity, xlo=1, xhi=5)

    show(gridplot([[el, el_c], [sp, sp_c], [hri, hri_c], [hr, hr_c], [pw, pw_c], [cd, hr_h]]))

    '''
    ## Activity Maps
    '''

    map = map_plot(400, 400, activity)
    m_el = map_intensity_signed(200, 200, activity, GRADE_PC, ranges=map, power=0.5)
    m_sp = map_intensity(200, 200, activity, SPEED_KMH, ranges=map, power=2)
    m_hr = map_intensity(200, 200, activity, HR_IMPULSE_10, ranges=map)
    m_pw = map_intensity(200, 200, activity, MED_POWER_ESTIMATE_W, ranges=map)
    show(row(map, gridplot([[m_el, m_sp], [m_hr, m_pw]], toolbar_location='right')))

    '''
    ## Activity Statistics
    '''

    '''
    Active time and distance exclude pauses.
    '''

    details[[ACTIVE_TIME, ACTIVE_DISTANCE]].dropna(). \
        transform({ACTIVE_TIME: format_seconds, ACTIVE_DISTANCE: format_metres})

    '''
    Climbs are auto-detected and shown only for the main activity. They are included in the elevation plot above.
    '''

    if present(details, CLIMB_TIME):
        display(transform(details.filter(like='Climb').dropna(),
                          {CLIMB_TIME: format_seconds, CLIMB_ELEVATION: format_metres,
                           CLIMB_DISTANCE: format_km, CLIMB_GRADIENT: format_percent,
                           CLIMB_POWER: format_watts, CLIMB_CATEGORY: lambda x: x}))

    '''
    ## Health and Fitness
    '''

    fitness, fatigue = like(FITNESS_D_ANY, health.columns), like(FATIGUE_D_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, TIME, fitness + fatigue, health, colours, alphas=alphas)
    xrange = ff.x_range if ff else None
    add_multi_line_at_index(ff, TIME, fitness + fatigue, health, colours, alphas=alphas, index=-1)
    atd = std_distance_time_plot(900, 200, health, x_range=ff.x_range)
    shr = multi_plot(900, 200, TIME, [DAILY_STEPS, REST_HR], health, ['grey', 'red'], alphas=[1, 0.5],
                     x_range=xrange, rescale=True, plotters=[bar_plotter(dt.timedelta(hours=20)), dot_plotter()])
    add_band(shr, TIME, LO_REST_HR, HI_REST_HR, health, 'red', alpha=0.1, y_range_name=REST_HR)
    add_curve(shr, TIME, REST_HR, health, color='red', y_range_name=REST_HR)
    show(column(ff, atd, shr))



from json import dumps
from logging import getLogger

from .database import add_statistics, add_enum_constant, set_constant, name_constant
from ..sql import ActivityGroup
from ..sql.types import short_cls
from ..stats.calculate.impulse import HRImpulse, ImpulseCalculator
from ..stats.calculate.response import Response, ResponseCalculator
from ..stats.names import FITNESS_D, FATIGUE_D, ALL
from ..stats.read.segment import SegmentReader

log = getLogger(__name__)


def add_impulse(s, c, activity_group):
    '''
    Add configuration for a fitness/fatigue impulse model based on HR zones.

    This adds:
    * HeartRateCalculator pipeline class to calculate HR zone and impulse
      * Impulse constant with parameters for impulse calculation
    * ImpulseCalculator pipeline class to calculate fitness and fatigue
      * Fitness constant with parameters for fitness calculation
      * Fatigue constant with parameters for fatigue calculation
    '''

    hr_impulse_name = name_constant('HRImpulse', activity_group)
    add_enum_constant(s, hr_impulse_name, HRImpulse,
                      {'gamma': 2.0, 'zero': 1, 'one': 6, 'max_secs': 60},
                      single=True, constraint=activity_group, description='''
Data needed to calculate the FF-model impulse from heart rate zones.
* Gamma is an exponent used to weight the relative importance of hard and easy efforts and should be around 1 (typical range 0.5 to 2).
* Zero defines the lowest heart zone used (so a value of 2 means that a HR zone of 2.0 or less contributes zero impulse).
* One defines the range of heart rate zones used (so a value of 6 means that a HR zone of 6.0 gives a unit impulse per time interval).
* Max_secs is the maximum gap between heart rate measurements (data that exceed this give no impulse).
Once the impulse is calculated it is summed with a decay to find fitness and fatigue
(see Fitness and Fatigue constants). 
''')
    add_statistics(s, ImpulseCalculator, c, owner_in=short_cls(SegmentReader), impulse_ref=hr_impulse_name,
                   activity_group_name=activity_group.name)


def add_responses(s, c, fitness=((42, 1, 1),), fatigue=((7, 1, 5),)):
    responses = []
    all = ActivityGroup.from_name(s, ALL)

    for days, start, scale in fitness:
        name = FITNESS_D % days
        constant = name_constant(name, all)
        responses.append(constant)
        log.debug(f'Adding fitness {name}')
        add_enum_constant(s, constant, Response,
                          {'src_owner': short_cls(ImpulseCalculator),
                           'dest_name': name, 'tau_days': days, 'start': start, 'scale': scale},
                          single=True, constraint=all, description=f'''
Data needed to calculate the FF-model fitness for {days} days.
* Src_owner is the process that generated the input data (ImpulseCalculator calculates the HR impulses).
* Dest_name is the statistic name where the results are stored.
* Tau_days is the time period (in days) over which the fitness decays.
* Start is the initial fitness value.
* Scale is an arbitrary scale factor (typically used so that fitness and fatigue have comparable values).
''')

    for days, start, scale in fatigue:
        name = FATIGUE_D % days
        constant = name_constant(name, all)
        responses.append(constant)
        log.debug(f'Adding fatigue {name}')
        add_enum_constant(s, constant, Response,
                          {'src_owner': short_cls(ImpulseCalculator),
                           'dest_name': name, 'tau_days': days, 'start': start, 'scale': scale},
                          single=True, constraint=all, description=f'''
Data needed to calculate the FF-model fatigue for {days} days.
* Src_owner is the process that generated the input data (ImpulseCalculator calculates the HR impulses).
* Dest_name is the statistic name where the results are stored.
* Tau_days is the time period (in days) over which the fitness decays.
* Start is the initial fitness value.
* Scale is an arbitrary scale factor (typically used so that fitness and fatigue have comparable values).
''')
    add_statistics(s, ResponseCalculator, c, owner_in=short_cls(ImpulseCalculator), responses_ref=responses)



from logging import getLogger
from sys import stdout

from .args import SUB_COMMAND, GROUP, ITEM, DATE, FORCE, COMPONENT, MODEL, STATISTICS, NAME, SHOW, CSV, \
    START, CHANGE, FINISH, DELETE, mm, UNDO, ALL, REBUILD, DUMP, KIT, CMD, VALUE, BASE
from ..diary.model import TYPE, UNITS
from ..lib import time_to_local_time, local_time_or_now, local_time_to_time, now, format_km, \
    is_local_time
from ..lib.date import format_minutes
from ..lib.tree import to_tree, to_csv
from ..sql import PipelineType
from ..sql.tables.kit import KitGroup, KitItem, KitComponent, KitModel, get_name, ADDED, EXPIRED, N, INDIVIDUAL
from ..sql.tables.source import Composite
from ..sql.types import long_cls
from ..stats.calculate.kit import KitCalculator
from ..stats.names import KM, S, _s
from ..stats.pipeline import run_pipeline

log = getLogger(__name__)


def kit(args, system, db, output=stdout):
    '''
## kit

Track equipment, including the lifetime of particular components.

    > ch2 kit new GROUP ITEM
    > ch2 kit change ITEM COMPONENT MODEL
    > ch2 kit statistics ITEM

For full details see `ch2 kit -h` and `ch2 kit SUBCOMMAND -h`.

### Examples

Note that in practice some commands that do 'important' changes to the database require `--force` for confirmation.

    > ch2 kit start bike cotic
    > ch2 kit change cotic chain sram --start
    # ... some months later ...
    > ch2 kit change cotic chain kmc
    # ... more time later ...
    > ch2 kit change cotic chain sram
    > ch2 kit statistics chain

This example will give statistics on how long (time, distance) different bikes chains lasted.

In addition, when importing activities, the `kit` variable must be defined.  So, for example:

    > ch2 activities -D kit=cotic **/*.fit

In this way the system knows what equipment was used in what activity.

Finally, statistics may be incorrect if the equipment is modified (because the correct use will not be
associated with each activity).  To recalculate use

    > ch2 kit rebuild

For running shoes you might simply track each item:

    > ch2 kit start shoe adidas
    # ... later ...
    > ch2 kit finish adidas
    > ch2 kit start shoe nike

Statistics for shoes:

    > ch2 kit statistic shoe

Names can be chosen at will (there is nothing hard-coded about 'bike', 'chain', 'cotic', etc),
but in general must be unique.  They can contain spaces if quoted.
    '''
    cmd, base = args[SUB_COMMAND], args[BASE]
    if cmd == REBUILD:
        rebuild(system, db, base)
    else:
        with db.session_context() as s:
            if cmd == START:
                start(s, args[GROUP], args[ITEM], args[DATE], args[FORCE])
            elif cmd == FINISH:
                finish(s, args[ITEM], args[DATE], args[FORCE])
            elif cmd == DELETE:
                delete(s, args[NAME], args[FORCE])
            elif cmd == CHANGE:
                change(s, args[ITEM], args[COMPONENT], args[MODEL], args[DATE], args[FORCE], args[START])
            elif cmd == UNDO:
                undo(s, args[ITEM], args[COMPONENT], args[MODEL], args[DATE], args[ALL])
            elif cmd == SHOW:
                show(s, args[NAME], args[DATE], csv=args[CSV], output=output)
            elif cmd == STATISTICS:
                statistics(s, args[NAME], csv=args[CSV], output=output)
            elif cmd == DUMP:
                dump(s, args[CMD])


def start(s, group, item, date, force):
    date = local_time_or_now(date)
    group_instance = KitGroup.get_or_add(s, group, force=force)
    item_instance = KitItem.add(s, group_instance, item, date)
    log.info(f'Started {group_instance.name} {item_instance.name} '
             f'at {time_to_local_time(item_instance.time_added(s))}')


def finish(s, item, date, force):
    date = local_time_or_now(date)
    get_name(s, item, classes=(KitItem,), require=True).finish(s, date, force)
    log.info(f'Finished {item}')


def delete(s, name, force):
    s.expunge_all()
    instance = get_name(s, name, classes=(KitGroup, KitItem), require=True)
    if isinstance(instance, KitGroup) and not force:
        raise Exception(f'Specify {mm(FORCE)} to delete group')
    s.delete(instance)
    s.flush()
    for component in s.query(KitComponent).all():
        component.delete_if_unused(s)
    Composite.clean(s)


def change(s, item, component, model, date, force, start):
    item_instance = KitItem.get(s, item)
    if start:
        if date:
            raise Exception(f'Do not provide a date with {mm(START)}')
        date = item_instance.time_added(s)
    else:
        date = local_time_or_now(date)
    component_instance = KitComponent.get_or_add(s, component, force)
    model_instance = KitModel.add(s, item_instance, component_instance, model, date)
    log.info(f'Changed {item_instance.name} {component_instance.name} {model_instance.name} '
             f'at {time_to_local_time(model_instance.time_added(s))}')


def undo(s, item, component, model, date, all):
    item_instance = KitItem.get(s, item)
    component_instance = KitComponent.get(s, component)
    if all:
        if date:
            raise Exception(f'Provide date or {mm(ALL)}, not both')
        model_instance = KitModel.get(s, item_instance, component_instance, model, None)
        while model_instance:
            model_instance.undo(s)
            model_instance = KitModel.get(s, item_instance, component_instance, model, None, require=False)
    else:
        model_instance = KitModel.get(s, item_instance, component_instance, model, local_time_or_now(date))
        model_instance.undo(s)
    component_instance.delete_if_unused(s)


def rebuild(system, db, base):
    run_pipeline(system, db, base, PipelineType.STATISTIC, force=True, like=[long_cls(KitCalculator)])


def show(s, name, date, csv=None, output=stdout):
    # this is complicated because both name and date are optional so if only one is
    # supplied we need to guess which from the format.
    if name is None:  # date must be None too, since it comes second on command line
        name, time = '*', now()
    elif date is None:
        if is_local_time(name):  # only one was supplied and it looke like a date
            name, time = '*', local_time_to_time(name)
        else:
            time = now()
    else:
        time = local_time_to_time(date)
    if name == '*':
        models = [group.to_model(s, time=time, depth=3) for group in s.query(KitGroup).order_by(KitGroup.name).all()]
    else:
        models = [get_name(s, name, classes=(KitGroup, KitItem), require=True).to_model(s, time=time, depth=3)]
    driver = to_csv if csv else to_tree
    format = to_label_name_dates_csv if csv else to_label_name_dates
    for model in models:
        for line in driver(model, format, model_children):
            print(line, file=output)


CHILDREN = {KitGroup.SIMPLE_NAME: _s(KitItem.SIMPLE_NAME),
            KitItem.SIMPLE_NAME: _s(KitComponent.SIMPLE_NAME),
            KitComponent.SIMPLE_NAME: _s(KitModel.SIMPLE_NAME)}


def model_children(model):
    if TYPE in model and model[TYPE] in CHILDREN and CHILDREN[model[TYPE]] in model:
        log.debug(f'Traversing from {model[TYPE]} to {CHILDREN[model[TYPE]]}')
        yield from model[CHILDREN[model[TYPE]]]


def to_label_name_dates(model):
    if ADDED in model:
        return f'{model[TYPE]}: {model[NAME]}  {model[ADDED]} - {model[EXPIRED] or ""}', None
    else:
        return f'{model[TYPE]}: {model[NAME]}', None


def to_label_name_dates_csv(model):
    if ADDED in model:
        added = time_to_local_time(model[ADDED])
        expired = time_to_local_time(model[EXPIRED]) if model[EXPIRED] else ''
    else:
        added, expired = '', ''
    return f'{q(model[TYPE])},{q(model[NAME])},{q(added)},{q(expired)}', None


def statistics(s, name, csv=False, output=stdout):
    if name:
        models = [get_name(s, name, require=True).to_model(s, depth=3, statistics=INDIVIDUAL)]
    else:
        models = [group.to_model(s, depth=3, statistics=INDIVIDUAL)
                  for group in s.query(KitGroup).order_by(KitGroup.name).all()]
    driver = to_csv if csv else to_tree
    format = to_stats_csv if csv else to_stats
    log.debug(models)
    for model in models:
        for line in driver(model, format, model_children):
            print(line, file=output)


def stats_children(model):
    names = [key for key in model.keys() if key not in (NAME, UNITS)]
    if model[UNITS] == KM:
        format = format_km
    elif model[UNITS] == S:
        format = lambda s: format_minutes(int(s))
    else:
        format = lambda x: x
    return [{NAME: N, VALUE: model[N]}] + [{NAME: name, VALUE: format(model[name])} for name in names if name != N]


def to_stats(model):
    if TYPE in model:
        label = f'{model[TYPE]}: {model[NAME]}'
        if STATISTICS in model:
            log.debug(f'Extracting statistics from {model[TYPE]}')
            return label, model[STATISTICS]
        else:
            log.debug('No statistics in model')
            return label, None
    elif VALUE not in model:
        log.debug('Formatting statistic')
        return f'{model[NAME]}', stats_children(model)
    else:
        # leaf
        return f'{model[NAME]}: {model[VALUE]}', None


def to_stats_csv(model):
    if TYPE in model:
        label = f'{q(model[TYPE])},{q(model[NAME])}'
        if STATISTICS in model:
            return label, model[STATISTICS]
        else:
            return label, None
    elif VALUE not in model:
        return f'{q(model[NAME])}', stats_children(model)
    else:
        return f'{q(model[NAME])},{q(model[VALUE])}', None


def q(name):
    name = str(name)
    if ' ' in name:
        return f"'{name}'"
    else:
        return name


def qd(time):
    return q(time_to_local_time(time))


def dump(s, cmd):
    if cmd is None: cmd = 'ch2'
    print('#!/bin/sh')
    print('# script generated by')
    print(f'# > {cmd} {KIT} {DUMP} {mm(CMD)} {q(cmd)}')
    groups = s.query(KitGroup).order_by(KitGroup.name).all()
    for group in groups:
        delete_group(s, cmd, group)
    for group in groups:
        dump_group(s, cmd, group)
    print(f'{cmd} {KIT} {REBUILD}')


def delete_group(s, cmd, group):
    print(f'{cmd} {KIT} {DELETE} {mm(FORCE)} {q(group.name)}')


def dump_group(s, cmd, group):
    for item in s.query(KitItem).filter(KitItem.group == group).all():
        dump_item(s, cmd, item)


def dump_item(s, cmd, item):
    print(f'{cmd} {KIT} {START} {mm(FORCE)} {q(item.group.name)} {q(item.name)} {qd(item.time_added(s))}')
    for model in s.query(KitModel).filter(KitModel.item == item).all():
        dump_model(s, cmd, item, model)


def dump_model(s, cmd, item, model):
    print(f'{cmd} {KIT} {CHANGE} {mm(FORCE)} {q(item.name)} {q(model.component.name)} {q(model.name)} '
          f'{qd(model.time_added(s))}')






# add flag fix to fix things

# check everything needed for power estimates
# (including weight)

# check constants defined

# check configure was run (version)

# check whether permanent directory present

# check whether file system ok (space?)

# check whether all files are scanned

# check whether all activity topics entries have an activity for the file hash




import datetime as dt

from bokeh.io import output_file
from bokeh.layouts import column
from bokeh.plotting import show

from ch2.data import *
from ch2.stats.names import _log
from ch2.jupyter.decorator import template


@template
def health():

    '''
    # Health
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')
    health = std_health_statistics(s)

    '''
    ## Health and Fitness
    '''

    output_file(filename='/dev/null')

    fitness, fatigue = like(FITNESS_D_ANY, health.columns), like(FATIGUE_D_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, TIME, fitness + fatigue, health, colours, alphas=alphas)
    add_multi_line_at_index(ff, TIME, fitness + fatigue, health, colours, alphas=alphas, index=-1)
    atd = std_distance_time_plot(900, 200, health, x_range=ff.x_range)
    shr = multi_plot(900, 200, TIME, [DAILY_STEPS, REST_HR], health, ['grey', 'red'], alphas=[1, 0.5],
                     x_range=ff.x_range, rescale=True, plotters=[bar_plotter(dt.timedelta(hours=20)), dot_plotter()])
    add_band(shr, TIME, LO_REST_HR, HI_REST_HR, health, 'red', alpha=0.1, y_range_name=REST_HR)
    add_curve(shr, TIME, REST_HR, health, color='red', y_range_name=REST_HR)
    show(column(ff, atd, shr))




from logging import getLogger

from ...lib import format_date, time_to_local_date, to_time
from ...lib.log import log_current_exception
from ...sql import KitGroup, KitComponent, KitItem, KitModel, StatisticJournalTimestamp, StatisticName, \
    StatisticJournalType
from ...sql.utils import add
from ...stats.names import KIT_ADDED, KIT_RETIRED

log = getLogger(__name__)


def upgrade_kit(record, old, new):
    if not kit_imported(record, new):
        try:
            with old.session_context() as old_s:
                with new.session_context() as new_s:
                    copy_components(record, old_s, old, new_s)
                    copy_tree(record, old_s, old, new_s)
        except Exception as e:
            log_current_exception()
            record.warning(f'Aborting kit import: {e}')


def kit_imported(record, new):
    with new.session_context() as new_s:
        return bool(new_s.query(KitGroup).count() + new_s.query(KitComponent).count())


def copy_components(record, old_s, old, new_s):
    component = old.meta.tables['kit_component']
    for old_component in old_s.query(component).all():
        add(new_s, KitComponent(name=old_component.name))
        record.info(f'Component {old_component.name}')


def copy_tree(record, old_s, old, new_s):
    group = old.meta.tables['kit_group']
    for old_group in old_s.query(group).all():
        new_group = add(new_s, KitGroup(name=old_group.name))
        record.info(f'Group {old_group.name}')
        item = old.meta.tables['kit_item']
        for old_item in old_s.query(item).filter(item.c.group_id == old_group.id).all():
            new_item = add(new_s, KitItem(name=old_item.name, group=new_group))
            record.info(f'Item {old_item.name}')
            copy_statistics(record, old_s, old, old_item, new_s, new_item)
            model = old.meta.tables['kit_model']
            component = old.meta.tables['kit_component']
            for old_model in old_s.query(model).filter(model.c.item_id == old_item.id).all():
                old_component = old_s.query(component).filter(component.c.id == old_model.component_id).one()
                new_component = new_s.query(KitComponent).filter(KitComponent.name == old_component.name).one()
                new_model = add(new_s, KitModel(name=old_model.name, item=new_item, component=new_component))
                record.info(f'Model {old_model.name}')
                copy_statistics(record, old_s, old, old_model, new_s, new_model)


def copy_statistics(record, old_s, old, old_source, new_s, new_source):
    statistic_name = old.meta.tables['statistic_name']
    statistic_journal = old.meta.tables['statistic_journal']
    statistic_journal_timestamp = old.meta.tables['statistic_journal_timestamp']
    for name in (KIT_ADDED, KIT_RETIRED):
        old_timestamp = old_s.query(statistic_journal). \
            join(statistic_journal_timestamp). \
            join(statistic_name). \
            filter(statistic_name.c.name == name,
                   statistic_journal.c.source_id == old_source.id).one_or_none()
        if old_timestamp:
            new_statistic_name = StatisticName.add_if_missing(new_s, name, StatisticJournalType.TIMESTAMP,
                                                              None, None, type(new_source), None)
            add(new_s, StatisticJournalTimestamp(source=new_source, time=old_timestamp.time,
                                                 statistic_name=new_statistic_name))
            date = format_date(time_to_local_date(to_time(old_timestamp.time)))
            record.info(f'Statistic {name} at {date} for {old_source.name}')




from bokeh.plotting import output_notebook, show

from ch2.data import *
from ch2.jupyter.decorator import template


@template
def some_activities(constraint):

    f'''
    # Some Activities: {constraint}

    This displays thumbnails of routes that match the query over statistics.  For example,

        Active Distance > 40 & Active Distance < 60

    will show all activities with a distance between 40 and 60 km.
    '''

    '''
    $contents
    '''

    '''
    ## Build Maps
    
    Loop over activities, retrieve data, and construct maps. 
    '''

    s = session('-v2')
    maps = [map_thumbnail(100, 120, data)
            for data in (activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                             ACTIVE_DISTANCE, TOTAL_CLIMB,
                                             activity_journal=aj)
                         for aj in constrained_activities(s, constraint))
            if len(data[SPHERICAL_MERCATOR_X].dropna()) > 10]
    print(f'Found {len(maps)} activities')

    '''
    ## Display Maps
    '''

    output_notebook()
    show(htile(maps, 8))





from logging import getLogger

from .climb import add_climb, CLIMB_CNAME
from .database import add_loader_support, add_activity_group, add_activities, Counter, add_statistics, add_diary, \
    name_constant, add_monitor, add_activity_constant, add_constant, add_diary_topic, add_diary_topic_field, \
    add_activity_topic_field
from .impulse import add_impulse
from .impulse import add_responses
from ..commands.args import base_system_path, DB_VERSION, PERMANENT
from ..commands.garmin import GARMIN_USER, GARMIN_PASSWORD
from ..diary.model import TYPE, EDIT, FLOAT, LO, HI, DP, SCORE
from ..lib.schedule import Schedule
from ..sql import DiaryTopicJournal, StatisticJournalType, ActivityTopicField, SystemConstant
from ..sql.types import short_cls
from ..srtm.file import SRTM1_DIR
from ..stats.calculate.achievement import AchievementCalculator
from ..stats.calculate.activity import ActivityCalculator
from ..stats.calculate.elevation import ElevationCalculator
from ..stats.calculate.kit import KitCalculator
from ..stats.calculate.monitor import MonitorCalculator
from ..stats.calculate.response import ResponseCalculator
from ..stats.calculate.segment import SegmentCalculator
from ..stats.calculate.summary import SummaryCalculator
from ..stats.display.achievement import AchievementDiary
from ..stats.display.activity import ActivityDiary
from ..stats.display.monitor import MonitorDiary
from ..stats.display.nearby import NearbyDiary
from ..stats.display.response import ResponseDiary
from ..stats.display.segment import SegmentDiary
from ..stats.names import SPORT_CYCLING, SPORT_RUNNING, SPORT_SWIMMING, SPORT_WALKING, FITNESS_D, FTHR, BPM, \
    FATIGUE_D, LATITUDE, DEG, LONGITUDE, HEART_RATE, SPEED, DISTANCE, KM, MS, ALTITUDE, CADENCE, RPM, M, ALL
from ..stats.read.monitor import MonitorReader
from ..stats.read.segment import SegmentReader

log = getLogger(__name__)

BROKEN = 'broken'

BIKE = 'Bike'
RUN = 'Run'
SWIM = 'Swim'
WALK = 'Walk'



class Config:
    '''
    A class-based approach so that we can easily modify the config for different profiles.
    '''

    def __init__(self, sys, base, no_diary=False):
        self._sys = sys
        self._base = base
        self._no_diary = no_diary
        self._activity_groups = {}

    def load(self, s):
        # hopefully you won't need to over-ride this, but instead one of the more specific methods
        self._pre(s)
        add_loader_support(s)  # required by standard statistics calculations
        self._load_specific_activity_groups(s)
        self._load_activity_group(s, ALL, 'All activities')  # referenced in FF calculations
        self._load_activities_pipeline(s, Counter())
        self._load_statistics_pipeline(s, Counter())
        self._load_diary_pipeline(s, Counter())
        self._load_monitor_pipeline(s, Counter())
        self._load_constants(s)
        if not self._no_diary:
            self._load_diary_topics(s, Counter())
            self._load_activity_topics(s, Counter())
            self._post_diary(s)
        self._post(s)

    def _pre(self, s):
        self._sys.set_constant(SystemConstant.DB_VERSION, DB_VERSION + ' ' + BROKEN)

    def _post_diary(self, s):
        pass

    def _post(self, s):
        self._sys.set_constant(SystemConstant.DB_VERSION, DB_VERSION, force=True)

    def _load_specific_activity_groups(self, s):
        # statistic rankings (best of month etc) are calculated per group
        self._load_activity_group(s, BIKE, 'Cycling activities')
        self._load_activity_group(s, RUN, 'Running activities')
        self._load_activity_group(s, SWIM, 'Swimming activities')
        self._load_activity_group(s, WALK, 'Walking activities')

    def _load_activity_group(self, s, name, description):
        log.debug(f'Loading activity group {name}')
        self._activity_groups[name] = add_activity_group(s, name, len(self._activity_groups), description=description)

    def _sport_to_activity(self):
        # sport_to_activity maps from the FIT sport field to the activity defined above.
        # more complex mapping is supported using kits (see acooke configuration),
        # but the top level must be a FIT field value for it to be recognized for monitor data
        # (which uses defaults only).
        return {SPORT_CYCLING: BIKE,
                SPORT_RUNNING: RUN,
                SPORT_SWIMMING: SWIM,
                SPORT_WALKING: WALK}

    def _all_groups_but_all(self):
        return [group for group in self._activity_groups.values() if group.name != ALL]

    def _record_to_db(self):
        # the mapping from FIT fields to database entries
        # you really don't want to alter this unless you know what you are doing...
        # todo - does this need to depend on activity group?
        return {'position_lat': (LATITUDE, DEG, StatisticJournalType.FLOAT),
                'position_long': (LONGITUDE, DEG, StatisticJournalType.FLOAT),
                'heart_rate': (HEART_RATE, BPM, StatisticJournalType.INTEGER),
                'enhanced_speed': (SPEED, MS, StatisticJournalType.FLOAT),
                'distance': (DISTANCE, KM, StatisticJournalType.FLOAT),
                'enhanced_altitude': (ALTITUDE, M, StatisticJournalType.FLOAT),
                'cadence': (CADENCE, RPM, StatisticJournalType.INTEGER)}

    def _load_activities_pipeline(self, s, c):
        sport_to_activity = self._sport_to_activity()
        record_to_db = self._record_to_db()
        add_activities(s, SegmentReader, c, owner_out=short_cls(SegmentReader),
                       sport_to_activity=sport_to_activity, record_to_db=record_to_db)

    def _ff_parameters(self):
        # each is a list of (days, start, scale) tuples.
        # for example, fitness with two different timescales:
        # fitness = ((42, 1, 1), (21, 1, 2))
        # https://andrewcooke.github.io/choochoo/ff-fitting
        fitness = ((42, 1, 1),)
        fatigue = ((7, 1, 5),)
        return fitness, fatigue

    def _impulse_groups(self):
        # activity groups that should have impulses calculated (ie that will contribute to FF statistics)
        return self._all_groups_but_all()

    def _load_power_statistics(self, s, c):
        # after elevation and before ff etc
        # defaults is none because configuration is complex
        pass

    def _load_ff_statistics(self, s, c, default_fthr=154):
        for group in self._impulse_groups():
            add_impulse(s, c, group)
            # we need a value here for various UI reasons.  might as well use my own value...
            add_activity_constant(s, group, FTHR, default_fthr,
                                  description=f'''
Heart rate (in bpm) at functional threshold for activities in the {group.name} group.

Your FTHR is the highest sustained heart rate you can maintain for long periods (an hour).
It is used to calculate how hard you are working (the Impulse) and, from that, 
your FF-model parameters (fitness and fatigue).
''',
                                  units=BPM, statistic_journal_type=StatisticJournalType.INTEGER)
        add_climb(s)  # default climb calculator
        fitness, fatigue = self._ff_parameters()
        add_responses(s, c, fitness=fitness, fatigue=fatigue)

    def _load_standard_statistics(self, s, c):
        add_statistics(s, ActivityCalculator, c, owner_in=short_cls(ResponseCalculator), climb=CLIMB_CNAME)
        add_statistics(s, SegmentCalculator, c, owner_in=short_cls(SegmentReader))
        add_statistics(s, MonitorCalculator, c, owner_in=short_cls(MonitorReader))
        add_statistics(s, KitCalculator, c, owner_in=short_cls(SegmentReader))

    def _load_summary_statistics(self, s, c):
        # need to call normalize here because schedule isn't a schedule type column,
        # but part of a kargs JSON blob.
        # also, add year first so that monthly doesn't get confused by extra stats range
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('x'))
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('y'))
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('m'))

    def _load_statistics_pipeline(self, s, c):
        # order is important here because some pipelines expect values created by others
        # this converts RAW_ELEVATION to ELEVATION, if needed
        add_statistics(s, ElevationCalculator, c, owner_in='[unused - data via activity_statistics]')
        self._load_power_statistics(s, c)
        self._load_ff_statistics(s, c)
        self._load_standard_statistics(s, c)
        self._load_summary_statistics(s, c)
        add_statistics(s, AchievementCalculator, c, owner_in=short_cls(SegmentReader))

    def _load_diary_pipeline(self, s, c):
        add_diary(s, MonitorDiary, c)
        # these tie-in to the constants used in add_impulse()
        fitness, fatigue = self._ff_parameters()
        all = self._activity_groups[ALL]
        add_diary(s, ResponseDiary, c,
                  fitness=[name_constant(FITNESS_D % days, all) for (days, _, _) in fitness],
                  fatigue=[name_constant(FATIGUE_D % days, all) for (days, _, _) in fatigue])
        add_diary(s, AchievementDiary, c)
        add_diary(s, ActivityDiary, c)
        add_diary(s, SegmentDiary, c)
        add_diary(s, NearbyDiary, c)

    def _load_monitor_pipeline(self, s, c):
        sport_to_activity = self._sport_to_activity()
        add_monitor(s, MonitorReader, c, sport_to_activity=sport_to_activity)

    def _load_constants(self, s):
        add_constant(s, SRTM1_DIR, base_system_path(self._base, version=PERMANENT, subdir='srtm1', create=False),
                     description='''
Directory containing STRM1 hgt files for elevations.

These data are used to give improved values when using GPS elevation.
You must create the directory and populate it with suitable files from http://dwtkns.com/srtm30m.
If the directory or files are missing the raw GPS elevation will be used.
This is noted as a warning in the logs (along with the name of the missing file).
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)
        add_constant(s, GARMIN_USER, None,
                     description='''
User for Garmin.
If set, monitor data (daily steps, heart rate) are downloaded from Garmin.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)
        add_constant(s, GARMIN_PASSWORD, None,
                     description='''
Password for Garmin.
This is stored as plaintext on the server (and visible here)
so do not use an important password that applies to many accounts.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)

    def _load_diary_topics(self, s, c):
        # the fields you see every day in the diary
        diary = add_diary_topic(s, 'Status', c)
        add_diary_topic_field(s, diary, 'Notes', c, StatisticJournalType.TEXT,
                              model={TYPE: EDIT})
        add_diary_topic_field(s, diary, 'Weight', c, StatisticJournalType.FLOAT,
                              units='kg', summary='[avg],[msr]',
                              model={TYPE: FLOAT, LO: 50, HI: 100, DP: 1})
        add_diary_topic_field(s, diary, 'Sleep', c, StatisticJournalType.FLOAT,
                              units='h', summary='[avg]',
                              model={TYPE: FLOAT, LO: 0, HI: 24, DP: 1})
        add_diary_topic_field(s, diary, 'Mood', c, StatisticJournalType.FLOAT,
                              summary='[avg]',
                              model={TYPE: SCORE})
        add_diary_topic_field(s, diary, 'Medication', c, StatisticJournalType.TEXT,
                              summary='[cnt]',
                              model={TYPE: EDIT})
        add_diary_topic_field(s, diary, 'Weather', c, StatisticJournalType.TEXT,
                              summary='[cnt]',
                              model={TYPE: EDIT})

    def _load_activity_topics(self, s, c):
        # the fields in the diary that are displayed for each activity
        for activity_group in self._all_groups_but_all():
            c = Counter()
            add_activity_topic_field(s, None, ActivityTopicField.NAME, c, StatisticJournalType.TEXT,
                                     activity_group, model={TYPE: EDIT})
            # note that these have empty toic parents because they are children of the entry itself
            if activity_group.name != SWIM:
                add_activity_topic_field(s, None, 'Route', c, StatisticJournalType.TEXT,
                                         activity_group, model={TYPE: EDIT})
            add_activity_topic_field(s, None, 'Notes', c, StatisticJournalType.TEXT,
                                     activity_group, model={TYPE: EDIT})

    def _load_sys(self, sys, s):
        # finally, update the timezone
        DiaryTopicJournal.check_tz(sys, s)





import numpy as np
from bokeh.io import show, output_file
from bokeh.plotting import figure

from ch2.data import *
from ch2.stats.read.monitor import MonitorReader
from ch2.jupyter.decorator import template


@template
def heart_rate(start, finish):

    f'''
    # Heart Rate: {start} - {finish}
    '''

    '''
    Extract and prepare data.
    '''

    bin_width = 1

    s = session('-v2')
    df = statistics(s, HEART_RATE, owner=MonitorReader, local_start=start, local_finish=finish)
    data = sorted(df[HEART_RATE])
    # take care here to get a fixed number of (integer) heart rates in each bin
    # this avoids aliasing effects.
    lo, hi = data[0] - 0.5, data[-1] + 0.5
    n = int(hi - lo + bin_width - 0.5) // bin_width
    hi = lo + n * bin_width
    hist, edges = np.histogram(data, density=True, bins=n, range=(lo, hi))
    y_max = max(hist)

    '''
    Plot histogram of heart rate and mark percentiles used to calculate rest value.
    
    The percentiles used are taken from the MonitorCalculator source (currently they are not configurable).
    It's not clear to me what the 'correct' values should be, but 10% seems to be reasonable, discarding possibly
    erroneous low counts but still giving a value from the bottom end of the range.
    '''

    output_file(filename='/dev/null')

    f = figure(title=f'Heart Rate: {start} - {finish}')
    f.xaxis.axis_label = 'Heart Rate / bpm'
    f.yaxis.axis_label = 'Measurement Density'
    f.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:], fill_color="grey", line_color="white", alpha=0.5)
    for pc in (5, 10, 15):
        x = data[int(len(data) * pc / 100)]
        f.line([x, x], [0, y_max], line_color='red', line_dash='dashed')
    show(f)




import re
from hashlib import sha1
from logging import getLogger
from os import stat
from pathlib import Path
from shutil import get_terminal_size

from sqlalchemy import desc

from .date import to_time
from ..sql.tables.file import FileScan, FileHash


log = getLogger(__name__)


def terminal_width(width=None):
    return get_terminal_size()[0] if width is None else width


def tui(command):
    def wrapper(*args, **kargs):
        return command(*args, **kargs)
    wrapper.tui = True
    wrapper.__doc__ = command.__doc__
    return wrapper


# https://stackoverflow.com/a/3431838
def file_hash(file_path):
    hash = sha1()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash.update(chunk)
    return hash.hexdigest()


def data_hash(data):
    hash = sha1()
    hash.update(data)
    return hash.hexdigest()


def modified_file_scans(s, paths, owner, force=False):

    modified = []

    for path in paths:

        # log.debug(f'Scanning {path}')
        last_modified = to_time(stat(path).st_mtime)
        hash = file_hash(path)
        file_scan_from_path = s.query(FileScan). \
            filter(FileScan.path == path,
                   FileScan.owner == owner).one_or_none()

        # get last scan and make sure it's up-to-date
        if file_scan_from_path:
            if hash != file_scan_from_path.file_hash.hash:
                log.warning('File at %s appears to have changed since last read on %s')
                file_scan_from_path.file_hash = FileHash.get_or_add(s, hash)
                file_scan_from_path.last_scan = 0.0
        else:
            file_scan_from_path = FileScan.add(s, path, owner, hash)
            s.flush()  # want this to appear in queries below

        # only look at hash if we are going to process anyway
        if force or last_modified > file_scan_from_path.last_scan:

            file_scan_from_hash = s.query(FileScan).\
                join(FileHash).\
                filter(FileHash.hash == hash,
                       FileScan.owner == owner).\
                order_by(desc(FileScan.last_scan)).limit(1).one()  # must exist as file_scan_from_path is a candidate
            if file_scan_from_hash.path != file_scan_from_path.path:
                log.warning('Ignoring duplicate file (details in debug log)')
                log.debug('%s' % file_scan_from_path.path)
                log.debug('%s' % file_scan_from_hash.path)
                # update the path to avoid triggering in future
                file_scan_from_path.last_scan = file_scan_from_hash.last_scan

            if force or last_modified > file_scan_from_hash.last_scan:
                modified.append(file_scan_from_hash)

    s.commit()
    return modified


def split_fit_path(path):
    # returns glob and kit
    pattern = re.compile(r'(.*\d\d\d\d-\d\d-\d\d.*)-([\w,]+).fit')
    match = pattern.match(path)
    if match:
        return match.group(1) + '*.fit', match.group(2)
    else:
        return path[:-4] + '*' + path[-4:], None


def touch(path):
    Path(path).touch()




from logging import getLogger

from werkzeug import Request, run_simple
from werkzeug.exceptions import HTTPException, BadRequest
from werkzeug.routing import Map, Rule
from werkzeug.wrappers.json import JSONMixin

from .json import JsonResponse
from .pages.analysis import Analysis
from .pages.configure import Configure
from .pages.diary import Diary
from .pages.jupyter import Jupyter
from .pages.kit import Kit
from .pages.upload import Upload
from .static import Static
from ..commands.args import mm, BASE, TUI, LOG, WEB, SERVICE, VERBOSITY, BIND, PORT, DEV, UPLOAD
from ..jupyter.server import JupyterController
from ..lib.log import log_current_exception
from ..lib.server import BaseController
from ..sql import SystemConstant

log = getLogger(__name__)


MAX_MSG = 1000

DATA = 'data'
REDIRECT = 'redirect'
ERROR = 'error'
BUSY = 'busy'
MESSAGE = 'message'
COMPLETE = 'complete'
PERCENT = 'percent'

GET = 'GET'
PUT = 'PUT'
POST = 'POST'


class JSONRequest(Request, JSONMixin):
    pass


class WebController(BaseController):

    def __init__(self, args, sys, db, max_retries=1, retry_secs=1):
        super().__init__(args, sys, WebServer, max_retries=max_retries, retry_secs=retry_secs)
        self.__bind = args[BIND] if BIND in args else None
        self.__port = args[PORT] if BIND in args else None
        self.__base = args[BASE]
        self.__dev = args[DEV]
        self.__sys = sys
        self.__db = db
        self.__jupyter = JupyterController(args, sys)

    def _build_cmd_and_log(self, ch2):
        log_name = 'web-service.log'
        cmd = f'{ch2} {mm(VERBOSITY)} {self._log_level} {mm(TUI)} {mm(LOG)} {log_name} {mm(BASE)} {self._base} ' \
              f'{WEB} {SERVICE} {mm(BIND)} {self.__bind} {mm(PORT)} {self.__port}'
        return cmd, log_name

    def _run(self):
        self._sys.set_constant(SystemConstant.WEB_URL, 'http://%s:%d' % (self.__bind, self.__port), force=True)
        run_simple(self.__bind, self.__port, WebServer(self.__sys, self.__db, self.__jupyter, self.__base),
                   use_debugger=self.__dev, use_reloader=self.__dev)

    def _cleanup(self):
        self._sys.delete_constant(SystemConstant.WEB_URL)

    def _status(self, running):
        if running:
            print(f'    {self.connection_url()}')
        print()

    def connection_url(self):
        return self._sys.get_constant(SystemConstant.WEB_URL, none=True)


def error(exception):
    def handler(*args, **kwargs):
        raise exception()
    return handler


class WebServer:

    def __init__(self, sys, db, jcontrol, base):
        self.__sys = sys
        self.__db = db

        analysis = Analysis()
        configure = Configure(sys, db, base)
        diary = Diary()
        jupyter = Jupyter(jcontrol)
        kit = Kit()
        static = Static('.static')
        upload = Upload(sys, db, base)

        self.url_map = Map([

            Rule('/api/analysis/parameters', endpoint=self.check(analysis.read_parameters), methods=(GET,)),

            Rule('/api/configure/profiles', endpoint=self.check(configure.read_profiles, config=False), methods=(GET,)),
            Rule('/api/configure/initial', endpoint=self.check(configure.write_profile, config=False), methods=(POST,)),
            Rule('/api/configure/delete', endpoint=self.check(configure.delete, config=False), methods=(POST,)),
            Rule('/api/configure/upgrade', endpoint=self.check(configure.read_upgrade), methods=(GET,)),
            Rule('/api/configure/upgrade', endpoint=self.check(configure.write_upgrade), methods=(POST,)),
            Rule('/api/configure/constants', endpoint=self.check(configure.read_constants), methods=(GET,)),
            Rule('/api/configure/constant', endpoint=self.check(configure.write_constant), methods=(PUT,)),

            Rule('/api/diary/neighbour-activities/<date>', endpoint=diary.read_neighbour_activities, methods=(GET,)),
            Rule('/api/diary/active-days/<month>', endpoint=diary.read_active_days, methods=(GET,)),
            Rule('/api/diary/active-months/<year>', endpoint=diary.read_active_months, methods=(GET,)),
            Rule('/api/diary/statistics', endpoint=self.check(diary.write_statistics), methods=(PUT,)),
            Rule('/api/diary/<date>', endpoint=self.check(diary.read_diary), methods=(GET,)),

            Rule('/api/jupyter/<template>', endpoint=jupyter, methods=(GET, )),

            Rule('/api/kit/edit', endpoint=self.check(kit.read_edit), methods=(GET, )),
            Rule('/api/kit/retire-item', endpoint=self.check(kit.write_retire_item), methods=(PUT,)),
            Rule('/api/kit/replace-model', endpoint=self.check(kit.write_replace_model), methods=(PUT,)),
            Rule('/api/kit/add-component', endpoint=self.check(kit.write_add_component), methods=(PUT,)),
            Rule('/api/kit/add-group', endpoint=self.check(kit.write_add_group), methods=(PUT,)),
            # used by upload
            Rule('/api/kit/items', endpoint=self.check(kit.read_items), methods=(GET,)),
            Rule('/api/kit/statistics', endpoint=self.check(kit.read_statistics), methods=(GET, )),
            Rule('/api/kit/<date>', endpoint=self.check(kit.read_snapshot), methods=(GET, )),

            Rule('/api/static/<path:path>', endpoint=static, methods=(GET, )),

            Rule('/api/upload', endpoint=self.check(upload), methods=(PUT, )),

            Rule('/api/busy', endpoint=self.read_busy, methods=(GET, )),
            Rule('/api/<path:_>', endpoint=error(BadRequest)),

            # ignore path and serve index.html
            Rule('/<path:_>', defaults={'path': 'index.html'}, endpoint=static, methods=(GET,)),
            Rule('/', defaults={'path': 'index.html'}, endpoint=static, methods=(GET,))
        ])

        self.__configure = configure

    def dispatch_request(self, request):
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            values.pop('_', None)
            with self.__db.session_context() as s:
                return endpoint(request, s, **values)
        except HTTPException as e:
            return e

    def wsgi_app(self, environ, start_response):
        request = JSONRequest(environ)
        response = self.dispatch_request(request)
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def get_busy(self):
        percent = self.__sys.get_percent(UPLOAD)
        if percent is None: percent = 100
        # the client uses the complete message when the problem has passed
        return {MESSAGE: 'Loading data and recalculating statistics.',
                COMPLETE: 'Data loaded and statistics updated.',
                PERCENT: percent}

    def read_busy(self, request, s):
        return JsonResponse({BUSY: self.get_busy()})

    def check(self, handler, config=True):

        def wrapper(request, s, *args, **kargs):
            if config and not self.__configure.is_configured():
                log.debug(f'Redirect (not configured)')
                return JsonResponse({REDIRECT: '/configure/initial'})
            busy = self.get_busy()
            if busy[PERCENT] is None or busy[PERCENT] == 100:
                try:
                    data = handler(request, s, *args, **kargs)
                    msg = f'Returning data: {data}'
                    if len(msg) > MAX_MSG:
                        msg = msg[:MAX_MSG-20] + ' ... ' + msg[-10:]
                    log.debug(msg)
                    return JsonResponse({DATA: data})
                except Exception as e:
                    log_current_exception()
                    # maybe some errors are redirects?
                    error = str(e).strip()
                    if not error.endswith('.'): error += '.'
                    log.debug(f'Returning error: {error}')
                    return JsonResponse({ERROR: error})
            else:
                log.debug(f'Returning busy: {busy}')
                return JsonResponse({BUSY: busy})

        return wrapper




from logging import getLogger

from .args import PATH, WARN
from ..fit.profile.profile import pickle_profile

log = getLogger(__name__)


def package_fit_profile(args, system, db):
    '''
## package-fit-profile

    > ch2 package-fit-profile data/sdk/Profile.xlsx

Parse the global profile and save the structures containing types and messages
to a pickle file that is distributed with this package.

This command is intended for internal use only.
    '''
    in_path, warn = args.file(PATH), args[WARN]
    pickle_profile(in_path, warn=warn)




from re import compile

'''
A parser generator takes a string as input and returns a generator that yields successive parses.
A parse is a (result, string) tuple where result is always a list and string is the remaining string (possibly empty).
'''


def literal(target):
    def _parser(string):
        if string.startswith(target):
            yield [target], string[len(target):]
    return _parser


def transform(parser, transform=lambda l: l):
    def _parser(string):
        for result, rest in parser(string):
            try:
                yield transform(result), rest
            except:  # allow filtering of inconsistent results (eg parse int)
                pass
    return _parser


def drop(parser):
    return transform(parser, lambda l: [])


# breadth first (depth first is inefficient)
def sequence(*parsers):
    def _recurse(parsers, results, string):
        if parsers:
            parser, parsers = parsers[0], parsers[1:]
            for result, rest in parser(string):
                yield from _recurse(parsers, results + result, rest)
        else:
            yield results, string
    def _parser(string):
        yield from _recurse(parsers, [], string)
    return _parser


def choice(*parsers):
    def _parser(string):
        for parser in parsers:
            yield from parser(string)
    return _parser


# only returns matching groups
def pattern(regexp):
    r = compile(regexp)
    def _parser(string):
        m = r.match(string)
        if m:
            yield list(m.groups()), string[m.end():]
    return _parser


class Recursive:

    def __init__(self):
        self._parser = None

    def __call__(self, *args, **kwargs):
        yield from self._parser(*args, **kwargs)

    def calls(self, parser):
        self._parser = parser


def exhaustive(parser):
    def _parser(string):
        for result, rest in parser(string):
            if not rest:
                yield result
    return _parser


def single(parser):
    def _parser(string):
        results = list(parser(string))
        if not results:
            raise Exception(f'Could not parse {string}')
        elif len(results) != 1:
            raise Exception(f'Ambiguous expression {string}')
        else:
            return results[0]
    return _parser




import asyncio
from logging import getLogger
from threading import Thread, Event
from time import sleep

from notebook.notebookapp import NotebookApp
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from ..commands.args import mm, NOTEBOOKS, JUPYTER, SERVICE, VERBOSITY, TUI, LOG, BASE, base_system_path
from ..lib.server import BaseController
from ..sql import SystemConstant

log = getLogger(__name__)


class JupyterServer(NotebookApp):

    def __init__(self, started, **kwargs):
        self._started = started
        super().__init__(**kwargs)

    @property
    def log_level(self):
        # this silences jupyter's logging
        # the existing mechanism only lets you set a value of 50, which means that "critical" messages
        # are still logged, and the decidedly non-critical usage instructions are printed.
        return 60

    def init_signal(self):
        log.debug('Skipping signal init')

    def start(self):
        self._started.set()
        super().start()


class JupyterController(BaseController):

    def __init__(self, args, sys, max_retries=5, retry_secs=3):
        super().__init__(args, sys, JupyterServer, max_retries=max_retries, retry_secs=retry_secs)

    def _status(self, running):
        if running:
            print(f'    {self.connection_url()}')
        print()

    def _build_cmd_and_log(self, ch2):
        log_name = 'jupyter-service.log'
        cmd = f'{ch2} {mm(VERBOSITY)} {self._log_level} {mm(TUI)} {mm(LOG)} {log_name} {mm(BASE)} {self._base} ' \
              f'{JUPYTER} {SERVICE}'
        return cmd, log_name

    def _cleanup(self):
        self._sys.delete_constant(SystemConstant.JUPYTER_URL)

    def connection_url(self):
        self.start()
        return self._sys.get_constant(SystemConstant.JUPYTER_URL)

    def base_dir(self):
        return self._base

    def _run(self):

        started = Event()

        def start():
            log.info('Starting Jupyter server in separate thread')
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
            notebook_dir = base_system_path(self._base, subdir=NOTEBOOKS)
            JupyterServer.launch_instance(['--notebook-dir', notebook_dir], started=started)

        t = Thread(target=start)
        t.daemon = True
        t.start()
        started.wait()  # set in JupyterServer.start() which is as late as we can get in startup
        log.debug('Separate thread started')

        while not hasattr(JupyterServer._instance, 'connection_url') or not JupyterServer._instance.connection_url:
            log.debug('Waiting for connection URL')
            sleep(1)

        self._sys.set_constant(SystemConstant.JUPYTER_URL, JupyterServer._instance.connection_url, force=True)

        log.info('Jupyter server started')
        while True:
            sleep(1)


__CONTROLLER_SINGLETON = None


def set_controller(controller):
    global __CONTROLLER_SINGLETON
    if __CONTROLLER_SINGLETON:
        raise Exception('Jupyter controller already set')
    __CONTROLLER_SINGLETON = controller
    return __CONTROLLER_SINGLETON


def get_controller():
    if not __CONTROLLER_SINGLETON:
        raise Exception('Jupyter controller not started')
    return __CONTROLLER_SINGLETON




from functools import lru_cache

import numpy as np
from scipy.interpolate import RectBivariateSpline

from .file import SRTM1_DIR, SAMPLES, ElevationSupport, elevation_from_constant, cached_file_reader


def spline_elevation_from_constant(s, dir_name=SRTM1_DIR, smooth=0):
    return elevation_from_constant(s, lambda dir: SplineElevation(dir, smooth), dir_name=dir_name)


class SplineElevation(ElevationSupport):

    def __init__(self, dir, smooth=0):
        super().__init__(dir, make_cached_spline_builder(smooth))

    def elevation(self, lat, lon):
        if self._dir:
            _, _, spline = self._lookup(lat, lon)
            return spline([lat], [lon])[0][0]  # it's a grid for all x and y, hence 2d
        else:
            return None


def make_cached_spline_builder(smooth):

    @lru_cache(4)  # 4 means our tests are quick (and should tile a local patch)
    def cached_spline_builder(dir, flat, flon):
        h = cached_file_reader(dir, flat, flon)
        x, y = np.linspace(flat, flat+1, SAMPLES), np.linspace(flon, flon+1, SAMPLES)
        # not 100% sure on the scaling of s but it seems to be related to sum of errors at all points
        # however, a scaling of SAMPLES * SAMPLES means that smooth=1 gives a numerical error, so add 10
        return RectBivariateSpline(x, y, h, s=smooth * SAMPLES * SAMPLES * 10)

    return cached_spline_builder



from logging import getLogger

from .args import SUB_COMMAND, SERVICE, START, STOP, SHOW, JUPYTER, LIST, PROGNAME, NAME, ARG, STATUS
from ..jupyter.server import set_controller, JupyterController
from ..jupyter.utils import templates, get_template

log = getLogger(__name__)


def jupyter(args, system, db):
    '''
## jupyter

    > ch2 jupyter show ...

Show the template in the browser, starting a background Jupyter server if necessary.

    > ch2 jupyter list

List the available templates and their arguments.

    > ch2 jupyter status

Indicate whether the background server is running or not.

    > ch2 jupyter stop

Stop the background server.
    '''
    cmd = args[SUB_COMMAND]
    if cmd == LIST:
        list()
    else:
        c = JupyterController(args, system)
        if cmd == STATUS:
            c.status()
        elif cmd == SHOW:
            set_controller(c)  # c is passed implicitly to template via global
            show(args)
        elif cmd == SERVICE:
            c.service()
        elif cmd == START:
            c.start(restart=True)
        elif cmd == STOP:
            c.stop()
        else:
            raise Exception(f'Unexpected command {cmd}')


def list():
    for name, (_, spec) in templates().items():
        args = ' '.join(spec.args)
        if spec.varargs:
            if args: args += ' '
            args += '*' + spec.varargs
        print(f'\n  {name}  {args}')
    print()


def check_params(params, spec):
    if len(params) < len(spec.args) or (not spec.varargs and len(params) > len(spec.args)):
        raise Exception(f'Received {len(params)} args but need {"at least " if spec.varargs else ""}'
                        f'{len(spec.args)} values (see {PROGNAME} {JUPYTER} {LIST})')
    return params


def show(args):
    name = args[NAME]
    params = args[ARG]
    try:
        fn, spec = get_template(name)
        params = check_params(params, spec)
        fn(*params)
    except KeyError:
        raise Exception(f'No template called {name} (see {PROGNAME} {JUPYTER} {LIST})')




from .constraint import constrained_activities
from .frame import df, session, statistics, statistic_quartiles, activity_statistics, std_activity_statistics, \
    std_health_statistics, nearby_activities, bookmarks, statistic_names, statistics, present, linear_resample_time, \
    groups_by_time, coallesce, transform, drop_empty, coallesce_like
from .power import fit_power
from .heart_rate import *
from .plot import col_to_boxstats, box_plot, line_plotter, dot_plotter, bar_plotter, add_climbs, multi_plot, \
    multi_dot_plot, multi_bar_plot, multi_line_plot, map_thumbnail, map_intensity, map_plot, histogram_plot, \
    cumulative_plot, htile, vtile, comparison_line_plot, map_intensity_signed, add_hr_zones, \
    add_multi_line_at_index, std_distance_time_plot, add_band, get_renderer, add_curve, add_climb_zones
from .lib import chisq, fit, inplace_decay
from .text import *
from ..stats.display.nearby import nearby_earlier, nearby_any_time
from ..stats.names import *

# avoid cleaning of imports
nearby_earlier, nearby_any_time, CLIMB_DISTANCE




from ..model import DB
from ...sql import StatisticJournal


def str_or_none(x):
    return x is None or isinstance(x, str)


def rewrite_db(model):
    if isinstance(model, list):
        return [rewrite_db(m) for m in model]
    else:
        if DB in model:
            db = model[DB]
            if isinstance(db, StatisticJournal):
                model[DB] = db.id
            else:  # links have tuples
                model[DB] = [x if str_or_none(x) else x.id for x in db]
        return model




from sqlalchemy import inspect

from ..lib.data import dict_to_attr


def tables(*classes):
    return dict_to_attr(dict((cls.__name__, inspect(cls).local_table) for cls in classes))


def add(s, instance):
    s.add(instance)
    return instance




from logging import getLogger
from time import sleep

from requests import HTTPError

from .args import DIR, USER, PASS, DATE, FORCE, base_system_path, PERMANENT, BASE
from ..fit.download.connect import GarminConnect
from ..lib import now, local_time_to_time, time_to_local_time
from ..lib.log import log_current_exception
from ..lib.utils import clean_path
from ..lib.workers import ProgressTree
from ..sql import Constant, SystemConstant
from ..stats.read.monitor import missing_dates

log = getLogger(__name__)

GARMIN_USER = 'Garmin.User'
GARMIN_PASSWORD = 'Garmin.Password'


def garmin(args, sys, db):
    '''
## garmin

    > ch2 garmin --user USER --pass PASSWORD DIR

Download recent monitor data to the given directory.

    > ch2 garmin --user USER --pass PASSWORD --date DATE DIR

Download monitor data for the given date.

Note that this cannot be used to download more than 10 days of data.
For bulk downloads use
https://www.garmin.com/en-US/account/datamanagement/
    '''
    dates = [args[DATE]] if args[DATE] else []
    dir = clean_path(DIR) if args[DIR] else None
    with db.session_context() as s:
        run_garmin(sys, s, dir=dir, base=args[BASE],
                   user=args[USER], password=args[PASS], dates=dates, force=args[FORCE])


def run_garmin(sys, s, dir=None, base=None, user=None, password=None, dates=None, force=False, progress=None):

    if not dates: dates = list(missing_dates(s, force=force))
    local_progress = ProgressTree(len(dates), parent=progress)
    if not (base or dir):
        raise Exception('Provide one of base or dir')

    try:
        if not dates:
            log.info('No missing data to download')
            return

        old_format = bool(dir)
        data_dir = dir or base_system_path(base, version=PERMANENT)
        user = user or Constant.get_single(s, GARMIN_USER)
        password = password or Constant.get_single(s, GARMIN_PASSWORD)

        last = sys.get_constant(SystemConstant.LAST_GARMIN, none=True)
        if last and (now() - local_time_to_time(last)).total_seconds() < 12 * 60 * 60:
            log.info(f'Too soon since previous call ({last}; 12 hours minimum)')
            return
        sys.set_constant(SystemConstant.LAST_GARMIN, time_to_local_time(now()), True)

        connect = GarminConnect(log_response=False)
        connect.login(user, password)

        for repeat, date in enumerate(dates):
            if repeat:
                sleep(1)
            log.info('Downloading data for %s' % date)
            try:
                with local_progress.increment_or_complete():
                    connect.get_monitoring_to_fit_file(date, data_dir, old_format=old_format)
            except HTTPError:
                log_current_exception(traceback=False)
                log.info('End of data')
                return

    finally:
        local_progress.complete()





from logging import getLogger

from sqlalchemy import Column, Text, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from ..support import Base
from ..types import Sort

log = getLogger(__name__)


class Achievement(Base):

    __tablename__ = 'achievement'

    id = Column(Integer, primary_key=True)
    text = Column(Text, nullable=False)
    sort = Column(Sort, nullable=False)
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal')
    UniqueConstraint(activity_journal_id, text)





from logging import getLogger
from os.path import join, exists
from shutil import rmtree

from .args import no, mm, SUB_COMMAND, CHECK, DATA, CONFIGURE, ACTIVITY_GROUPS, LIST, PROFILE, DIARY, DELETE, FORCE, \
    BASE
from .help import Markdown
from ..config import ActivityGroup
from ..config.utils import profiles, get_profile
from ..lib.utils import slow_warning
from ..sql import SystemConstant
from ..sql.tables.statistic import StatisticName, StatisticJournal

log = getLogger(__name__)


def configure(args, sys, db):
    '''
## configure

    > ch2 configure load default

Generate a simple initial configuration.

Please see the documentation at http://andrewcooke.github.io/choochoo - you have a lot more options!

    > ch2 configure check --no-config --no-data

Check that the current database is empty.
    '''
    action = args[SUB_COMMAND]
    if action == CHECK:
        check(db, args[no(CONFIGURE)], args[no(DATA)], args[no(ACTIVITY_GROUPS)])
    elif action == LIST:
        list()
    elif action == DELETE:
        delete(sys, args.system_path(), args[FORCE])
    else:
        with db.session_context() as s:
            load(sys, s, args[BASE], args[no(DIARY)], args[PROFILE])


def check(db, config, data, activity_groups):
    with db.session_context() as s:
        if config:
            if s.query(StatisticName).count():
                raise Exception('The database already contains StatisticName entries')
        if data:
            if s.query(StatisticJournal).count():
                raise Exception('The database already contains StatisticJournal entries')
        if activity_groups:
            if s.query(ActivityGroup).count():
                raise Exception('The database already contains ActivityGroup entries')


def list():
    fmt = Markdown()
    for name in profiles():
        fn, spec = get_profile(name)
        if fn.__doc__:
            fmt.print(fn.__doc__)
        else:
            print(f' ## {name} - lacks docstring\n')


def load(sys, s, base, no_diary, profile):
    version = sys.get_constant(SystemConstant.DB_VERSION, none=True)
    if version:
        raise Exception(f'System already configured with version {version}')
    fn, spec = get_profile(profile)
    log.info(f'Loading profile {profile}')
    fn(sys, s, base, no_diary)
    log.info(f'Profile {profile} loaded successfully')


def delete(sys, path, force):
    if not force:
        raise Exception(f'If you really want to delete all your data from {base} add {mm(FORCE)}')
    data = join(path, DATA)
    log.debug(f'Checking {data}')
    if not exists(data):
        raise Exception(f'The directory {data} does not exist')
    version = sys.get_constant(SystemConstant.DB_VERSION, none=True)
    if not version:
        log.info(f'Deleting unconfigured system')
    slow_warning(f'Deleting {path}')
    rmtree(path)




from glob import glob
from logging import getLogger
from os.path import basename, join, exists

from ch2.commands.args import FIELDS, TABLES, GREP
from ch2.fit.format.read import filtered_records
from ch2.fit.format.records import no_names, append_units, no_bad_values, fix_degrees, chain
from ch2.fit.profile.fields import DynamicField
from ch2.fit.profile.profile import read_external_profile, read_fit
from ch2.fit.summary import summarize, summarize_csv, summarize_tables
from ch2.lib.tests import OutputMixin, HEX_ADDRESS, EXC_HDR_CHK, sub_extn, EXC_FLD, sub_dir, RNM_UNKNOWN, ROUND_DISTANCE
from tests import LogTestCase

log = getLogger(__name__)


class TestFit(LogTestCase, OutputMixin):

    # some of these tests depend on the width of the window they are run in!  :(
    # (use 80 chars until fixed)

    def setUp(self):
        self.test_dir = 'data/test'
        self.profile_path = 'data/sdk/Profile.xlsx'

    def test_profile(self):
        nlog, types, messages = read_external_profile(self.profile_path)
        cen = types.profile_to_type('carry_exercise_name')
        self.assertEqual(cen.profile_to_internal('farmers_walk'), 1)
        session = messages.profile_to_message('session')
        field = session.profile_to_field('total_cycles')
        self.assertIsInstance(field, DynamicField)
        for name in field.references:
            self.assertEqual(name, 'sport')
        workout_step = messages.profile_to_message('workout_step')
        field = workout_step.number_to_field(4)
        self.assertEqual(field.name, 'target_value')
        fields = ','.join(sorted(field.references))
        self.assertEqual(fields, 'duration_type,target_type')

    def test_decode(self):
        types, messages, records = \
            filtered_records(read_fit(join(self.test_dir, 'source/personal/2018-07-26-rec.fit')),
                             profile_path=self.profile_path)
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_decode'),
                                  filters=[HEX_ADDRESS]) as output:
            for _, _, record in records:
                print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)),
                      file=output)

    def test_dump(self):
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_dump')) as output:
            summarize(FIELDS, read_fit(join(self.test_dir, 'source/personal/2018-07-30-rec.fit')),
                      profile_path=self.profile_path, width=80, output=output)

    def test_developer(self):
        with self.assertTextMatch(join(self.test_dir, 'target/sdk/TestFit.test_developer')) as output:
            summarize(FIELDS, read_fit(join(self.test_dir, 'source/sdk/DeveloperData.fit')),
                      profile_path=self.profile_path, width=80, output=output)

    def test_personal(self):
        for fit_file in glob(join(self.test_dir, 'source/personal/*.fit')):
            file_name = basename(fit_file)
            with self.assertTextMatch(
                    join(self.test_dir, 'target/personal/TestFit.test_personal:' + file_name)) as output:
                summarize_tables(read_fit(fit_file), width=80, output=output,
                                 profile_path=self.profile_path)

    def test_timestamp_16(self):
        types, messages, records = \
            filtered_records(read_fit(join(self.test_dir, 'source/personal/andrew@acooke.org_24755630065.fit')),
                             profile_path=self.profile_path)
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_timestamp_16'),
                                  filters=[HEX_ADDRESS]) as output:
            for _, _, record in records:
                if record.name == 'monitoring':
                    print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)),
                          file=output)

    def standard_csv(self, fit_path, csv_path, filters=None):
        if filters is None: filters = []
        if EXC_HDR_CHK not in filters: filters = [EXC_HDR_CHK] + filters
        with self.assertCSVMatch(csv_path, filters=filters) as output:
            summarize_csv(read_fit(fit_path), profile_path=self.profile_path,
                          warn=True, output=output, internal=True)

    def standard_csv_dir(self, dir, fit_pattern, exclude=None, filters=None):
        for source_fit_path in glob(join(self.test_dir, 'source', dir, fit_pattern)):
            if not exclude or basename(source_fit_path) not in exclude:
                source_csv_path = sub_extn(source_fit_path, 'csv')
                if exists(source_csv_path):
                    self.standard_csv(source_fit_path, source_csv_path, filters=filters)
                else:
                    log.warning('Could not find %s' % source_csv_path)
                target_csv_path = sub_dir(source_csv_path, 'target', 2)
                if exists(target_csv_path):
                    self.standard_csv(source_fit_path, target_csv_path, filters=filters)
                else:
                    log.warning('Could not find %s' % target_csv_path)

    def standard_dmp(self, source, target, format, filters=None):
        with self.assertTextMatch(target, filters=filters) as output:
            summarize(format, read_fit(source),
                      warn=True, profile_path=self.profile_path, output=output)

    def standard_dmp_dir(self, dir, fit_pattern, format, exclude=None, filters=None):
        for source_fit_path in glob(join(self.test_dir, 'source', dir, fit_pattern)):
            if not exclude or basename(source_fit_path) not in exclude:
                target_dmp_path = sub_extn(sub_dir(source_fit_path, 'target', 2), format[:3])
                self.standard_dmp(source_fit_path, target_dmp_path, format, filters=filters)

    def test_sdk_csv(self):
        self.standard_csv_dir('sdk', '*.fit', exclude='Activity.fit')
        # afaict it should be 0, which is mapped by the type.  the value in the CSV makes no sense.
        self.standard_csv_dir('sdk', 'Activity.fit', filters=[EXC_FLD('timer_trigger')])

    def test_personal_csv(self):
        self.standard_csv_dir('personal', '*.fit', filters=[RNM_UNKNOWN])

    # these are tests only for consistency across releases, not correctness
    def test_personal_tab(self):
        self.standard_dmp_dir('personal', '*.fit', TABLES)

    # these depend on the output from test_fix_fit, so will fail if those change
    def test_pyfitparse_tab(self):
        self.standard_dmp_dir('python-fitparse', '*.fit', TABLES,
                              exclude=['activity-unexpected-eof.fit',  # data size incorrect
                                       'activity-settings-nodata.fit',   # data size incorrect
                                       'activity-filecrc.fit',  # bad checksum
                                       'activity-activity-filecrc.fit',  # data size incorrect
                                       'activity-settings-corruptheader.fit',  # data size incorrect
                                       'event_timestamp.fit',  # data size incorrect
                                       'activity-settings.fit',  # data size incorrect
                                       ])

    # these are tests only for consistency across releases, not correctness
    def test_pyfitparse_fix_tab(self):
        self.standard_dmp_dir('python-fitparse-fix', '*.fit', TABLES)

    def test_pyfitparse_csv(self):
        self.standard_csv_dir('python-fitparse', '*.fit', filters=[RNM_UNKNOWN, ROUND_DISTANCE])

    def test_other_csv(self):
        self.standard_csv_dir('other', '*.fit', filters=[RNM_UNKNOWN, ROUND_DISTANCE])

    def test_grep(self):
        data = read_fit(join(self.test_dir, 'source/personal/2018-07-26-rec.fit'))
        summarize(GREP, data, grep=['.*:.*speed>10'])

    def test_python(self):
        # for other-projects.md
        from ch2.fit.profile.profile import read_fit, read_profile
        from ch2.fit.format.records import fix_degrees, no_units
        from ch2.fit.format.read import parse_data

        data = read_fit('data/test/source/personal/2018-07-26-rec.fit')
        types, messages = read_profile()
        state, tokens = parse_data(data, types, messages)

        LAT, LONG = 'position_lat', 'position_long'
        positions = []

        for offset, token in tokens:
            record = token.parse_token().as_dict(no_units, fix_degrees)
            if record.name == 'record':
                positions.append((record.data[LAT][0], record.data[LONG][0]))

        print('Read %s positions' % len(positions))

        self.assertAlmostEqual(positions[0][0], -33.42, places=1)
        self.assertAlmostEqual(positions[0][1], -70.61, places=1)




from .file import SRTM1_DIR, SAMPLES, ElevationSupport, elevation_from_constant


def bilinear_elevation_from_constant(s, dir_name=SRTM1_DIR):
    return elevation_from_constant(s, BilinearElevation, dir_name=dir_name)


class BilinearElevation(ElevationSupport):
    '''
    Provide elevation data from the files in `dir` which should be hgt files with standard naming,
    either zipped or unzipped, downloaded from http://dwtkns.com/srtm30m/.

    You can download *all* data by going to https://search.earthdata.nasa.gov/search?q=srtm%20v3 and
    selecting 'NASA Shuttle Radar Topography Mission Global 1 arc second V003'.  That will lead to
    a download page that provides a further links to a list of links and a download script.  Running
    the download script downloads the links.  You need to register.

    All coords are "GPS coords" afaict.

    If dir is None then None will be return as elevation for all queries.

    If dir is not None and a file is missing for a particular lat/lon then an exception is raised.

    Elevations are bilinear interpolated from the surrounding arcsec grid.
    '''

    def elevation(self, lat, lon):
        if self._dir:
            flat, flon, h = self._lookup(lat, lon)
            x = (lon - flon) * (SAMPLES - 1)
            y = (lat - flat) * (SAMPLES - 1)  # -1 because weird inclusive-at-each-side tiling
            i, j = int(x), int(y)
            # bilinear
            # it's ok to use +1 blindly here because we're never on the top/right cells or we'd be in
            # a different tile.
            k = y - j
            h0 = h[j, i] * (1-k) + h[j+1, i] * k
            h1 = h[j, i+1] * (1-k) + h[j+1, i+1] * k
            k = x - i
            return h0 * (1-k) + h1 * k
        else:
            return None




import datetime as dt
from logging import getLogger

import numpy as np
from pandas import DataFrame, Series
from scipy import optimize

from ch2.data.lib import decay_params
from . import inplace_decay
from ..stats.names import FITNESS_D_ANY, FATIGUE_D_ANY, like, _delta

log = getLogger(__name__)

IMPULSE_3600 = 'Impulse / 3600s'
RESPONSE = 'Response'
LOG10_PERIOD, LOG10_START = 0, 1


# Almost everything below uses Series, not DataFrame
# Currently model params are log10_period (hours) and log10_start (initial FF value)


def sum_to_hour(source, column):
    data = source.resample('1h', label='right').sum()
    data = data.loc[:, [column]]
    data.rename(columns={column: IMPULSE_3600}, inplace=True)
    initial = DataFrame({IMPULSE_3600: 0}, index=[data.index[0] - dt.timedelta(hours=1)])
    data = initial.append(data, sort=True)
    return data


def calc_response(data, params):
    # data is a DataFrame mainly because that's what inplace_decay takes
    response = data.rename(columns={IMPULSE_3600: RESPONSE})  # copy
    period = 10 ** params[LOG10_PERIOD]
    if len(params) > LOG10_START:
        decay, alpha = decay_params(period)
        # correct by alpha so that the value matches the plot (see decay_inplace)
        start = 10 ** params[LOG10_START] * alpha
        # we replace the very first value which is strictly cheating, but there are so many...
        response.at[response.index[0], RESPONSE] = start
    inplace_decay(response, RESPONSE, period)
    return response[RESPONSE]


def restrict_response(response, performances):
    # we're only interested in the FF model at the times that correspond to performance measurements
    return [response.reindex(index=performance.index, method='nearest')
            for performance in performances]


def calc_observed(measureds, performances):
    # this is a bit tricky.  we don't know the exact relationship between 'fitness' and how fast we are.
    # all we know is that it should track changes in a useful manner.
    # so there's some arbitrary transform, which could quite easily be different for different routes
    # (no reason 'fitness' should numerically predict the speed on *all* routes, obviously).
    # so we assume that there's an arbitrary linear transform (scaling and offset) from measured speed
    # (or whatever 'performance' is) to 'fitness'.
    # we could fit for the best transform as we fit for the period, but it's more efficient to calculate
    # the 'best' transform for a given period.  this is more efficient because it's equivalent to line fitting
    # (that linear transform is y = ax + b).  so we can call the (fast) line fitting routine in numpy.
    return [Series(np.polyval(np.polyfit(performance, measured, 1), performance),
                   performance.index)
            for measured, performance in zip(measureds, performances)]


def calc_residuals(models, observations, method='L1'):
    # -ve when observation less than model
    if method == 'L1':
        # use L1 scaled by amplitude to be reasonably robust.
        return [(observation - model) / model.clip(lower=1e-6)
                for model, observation in zip(models, observations)]
    elif method == 'L2':
        # something like chisq
        return [(np.sign(observation - model) * (observation - model).pow(2) / model.clip(lower=1e-6))
                for model, observation in zip(models, observations)]
    else:
        raise Exception(f'Unknown method ({method}) - use L1 or L2')


def calc_cost(model, observation, method='L1'):
    residuals = calc_residuals(model, observation, method=method)
    return sum(sum(np.abs(residual)) for residual in residuals)


def worst_residual(residuals, threshold=None):
    i, t, worst = None, None, None
    for index, residual in enumerate(residuals):
        bad = max(residual)
        if bad > threshold and (worst is None or bad > worst):
            worst, i, t = bad, index, residual.idxmax()
    if worst:
        log.debug(f'Worst residual magnitude {worst} at {t}')
    else:
        log.debug('No worst residual found')
    return i, t, worst


def fix_residuals(residuals, threshold=None):
    try:
        pos, neg = np.abs(threshold)
        log.debug(f'Scaling residuals to weight lower by {pos}:{neg}')
        for residual in residuals:
            residual.loc[residual < 0] = residual.loc[residual < 0].abs() * pos / neg
        return residuals, pos
    except TypeError:
        return [residual.abs() for residual in residuals], 0 if threshold is None else threshold


def reject_worst_inplace(params, data, performances, method='L1', threshold=None):
    response = calc_response(data, params)
    measureds = restrict_response(response, performances)
    predicteds = calc_observed(measureds, performances)
    residuals = calc_residuals(measureds, predicteds, method=method)
    residuals, threshold = fix_residuals(residuals, threshold=threshold)
    i, t, worst = worst_residual(residuals, threshold=threshold)
    if t:
        print(f'Dropping {worst} at {t}')
        performances[i].drop(index=t, inplace=True)
    return i, t


def fmt_params(params):
    text = f'Period {10 ** params[0] / 24:.1f}'
    if len(params) > 1:
        text += f'; Start {10 ** params[1]:.1f}'
    return text


def fit_ff_params(data, params, performances, method='L1', max_reject=0, threshold=None, **kargs):
    # data should be a DataFrame with an IMPULSE3600 entry
    # performances should be Series
    # threshold can be None, a value, or a pair.  a pair gives +ve and -ve thresholds

    result, rejected = None, []

    def cost(params):
        response = calc_response(data, params)
        restricted = restrict_response(response, performances)
        observed = calc_observed(restricted, performances)
        return calc_cost(restricted, observed, method=method)

    while True:
        result = optimize.minimize(cost, params, **kargs)
        print(fmt_params(result.x))
        print(f'Currently have {len(rejected)} dropped points (max {max_reject})')
        if len(rejected) >= max_reject:
            return result, rejected
        i, t = reject_worst_inplace(params, data, performances, method=method, threshold=threshold)
        if t is None:
            return result, rejected
        rejected.append((i, t))


def response_stats(df):
    stats = {}
    for pattern in FITNESS_D_ANY, FATIGUE_D_ANY:
        for name in like(pattern, df.columns):
            stats[_delta(name)] = df[name][-1] - df[name][0]
    return stats




from scipy.interpolate import UnivariateSpline

from .frame import present
from ..stats.names import ELEVATION, DISTANCE, RAW_ELEVATION, GRADE


def fix_elevation(df, smooth=4):
    if not present(df, ELEVATION):
        unique = df.loc[~df[DISTANCE].isna() & ~df[RAW_ELEVATION].isna(),
                        [DISTANCE, RAW_ELEVATION]].drop_duplicates(DISTANCE)
        # the smoothing factor is from eyeballing results only.  maybe it should be a parameter.
        # it seems better to smooth along the route rather that smooth the terrain model since
        # 1 - we expect the route to be smoother than the terrain in general (roads / tracks)
        # 2 - smoothing the 2d terrain is difficult to control and can give spikes
        # 3 - we better handle errors from mismatches between terrain model and position
        #     (think hairpin bends going up a mountainside)
        # the main drawbacks are
        # 1 - speed on loading
        # 2 - no guarantee of consistency between routes (or even on the same routine retracing a path)
        spline = UnivariateSpline(unique[DISTANCE], unique[RAW_ELEVATION], s=len(unique) * smooth)
        df[ELEVATION] = spline(df[DISTANCE])
        df[GRADE] = (spline.derivative()(df[DISTANCE]) * 100)  # two step to use rolling from pandas
        df[GRADE] = df[GRADE].rolling(5, center=True).median().ffill().bfill()
        # avoid extrapolation / interpolation
        df.loc[df[RAW_ELEVATION].isna(), [ELEVATION]] = None
    return df



from logging import getLogger

from werkzeug.utils import redirect

from ...jupyter.load import create_notebook
from ...jupyter.utils import get_template

log = getLogger(__name__)


class Jupyter:

    def __init__(self, controller):
        self.__controller = controller

    def __call__(self, request, s, template):
        log.info(f'Attempting to display template {template}')
        fn, spec = get_template(template)
        args = [request.args[arg] for arg in spec.args]  # order
        log.debug(f'Template args: {args}')
        name = create_notebook(fn, self.__controller.base_dir(), args, {})
        url = f'{self.__controller.connection_url()}tree/{name}'
        log.debug(f'Redirecting to {url}')
        return redirect(url)




from logging import getLogger

from .args import PATH, SUB_COMMAND, AFTER_BYTES, LIMIT_BYTES, AFTER_RECORDS, LIMIT_RECORDS, NAME, WIDTH, GREP, \
    RECORDS, ALL_FIELDS, INTERNAL, ALL_MESSAGES, MESSAGE, FIELD, VALIDATE, MAX_DELTA_T, no, WARN, TABLES, PATTERN, \
    COMPACT, CONTEXT, NOT, MATCH, CSV, TOKENS, FIELDS
from ..fit.profile.profile import read_fit
from ..fit.summary import summarize_records, summarize_tables, summarize_grep, summarize_csv, summarize_tokens, \
    summarize_fields
from ..lib.io import terminal_width

log = getLogger(__name__)


def fit(args, system, db):
    '''
## fit

    > ch2 fit SUB-COMMAND PATH [PATH ...]

Print the contents of fit files.

The format and details displayed are selected by the sub-command: records, tables, messages, fields, csv
and grep (the last requiring patterns to match against).

For a list of sub-commands options see `ch2 fit -h`.

For options for a particular sub-command see `ch2 fit sub-command -h`.

Note: When using bash use `shopt -s globstar` to enable ** globbing.

### Examples

    > ch2 -v 0 fit records ride.fit

Will print the contents of the file to stdout (use `-v 0` to suppress logging
or redirect stderr elsewhere).

    > ch2 -v 0 fit grep -p '.*:sport=cycling' --match 0 --name directory/**/*.fit

Will list file names that contain cycling data.

    > ch2 fit grep -p PATTERN -- FILE

You may need a `--` between patterns and file paths so that the argument parser can decide where patterns
finish and paths start.
    '''

    format = args[SUB_COMMAND]
    after_bytes = args[AFTER_BYTES]
    limit_bytes = args[LIMIT_BYTES]
    after_records = args[AFTER_RECORDS]
    limit_records = args[LIMIT_RECORDS]
    warn = args[WARN]
    no_validate = args[no(VALIDATE)]
    max_delta_t = args[MAX_DELTA_T]

    # todo - can this be handled by argparse?
    if (after_records or limit_records != -1) and (after_bytes or limit_bytes != -1):
        raise Exception('Constrain either records or bytes, not both')

    for file_path in args[PATH]:

        name_file = file_path if args[NAME] else None
        if name_file and format != GREP:
            print()
            print(name_file)

        data = read_fit(file_path)

        if format == RECORDS:
            summarize_records(data,
                              all_fields=args[ALL_FIELDS], all_messages=args[ALL_MESSAGES],
                              internal=args[INTERNAL], after_bytes=after_bytes, limit_bytes=limit_bytes,
                              after_records=after_records, limit_records=limit_records,
                              record_names=args[MESSAGE], field_names=args[FIELD],
                              warn=warn, no_validate=no_validate, max_delta_t=max_delta_t,
                              width=args[WIDTH] or terminal_width())
        elif format == TABLES:
            summarize_tables(data,
                             all_fields=args[ALL_FIELDS], all_messages=args[ALL_MESSAGES],
                             internal=args[INTERNAL], after_bytes=after_bytes, limit_bytes=limit_bytes,
                             after_records=after_records, limit_records=limit_records,
                             record_names=args[MESSAGE], field_names=args[FIELD],
                             warn=warn, no_validate=no_validate, max_delta_t=max_delta_t,
                             width=args[WIDTH] or terminal_width())
        elif format == CSV:
            summarize_csv(data,
                          internal=args[INTERNAL], after_bytes=after_bytes, limit_bytes=limit_bytes,
                          after_records=after_records, limit_records=limit_records,
                          record_names=args[MESSAGE], field_names=args[FIELD],
                          warn=warn, max_delta_t=max_delta_t)
        elif format == GREP:
            summarize_grep(data, args[PATTERN],
                           after_bytes=after_bytes, limit_bytes=limit_bytes,
                           after_records=after_records, limit_records=limit_records,
                           warn=warn, no_validate=no_validate, max_delta_t=max_delta_t,
                           width=args[WIDTH] or terminal_width(),
                           name_file=name_file, match=args[MATCH], compact=args[COMPACT],
                           context=args[CONTEXT], invert=args[NOT])
        elif format == TOKENS:
            summarize_tokens(data,
                             after_bytes=after_bytes, limit_bytes=limit_bytes,
                             after_records=after_records, limit_records=limit_records,
                             warn=warn, no_validate=no_validate, max_delta_t=max_delta_t)
        elif format == FIELDS:
            summarize_fields(data,
                             after_bytes=after_bytes, limit_bytes=limit_bytes,
                             after_records=after_records, limit_records=limit_records,
                             warn=warn, no_validate=no_validate, max_delta_t=max_delta_t)
        else:
            raise Exception('Bad format: %s' % format)
